from __future__ import annotations
import asyncio
import base64
from copy import copy
from functools import reduce
import logging
from time import time, sleep
from typing import (
    cast, List, Dict, Any, Optional, Set, Sequence, Tuple, TypeVar, Iterator)
from typing_extensions import Final
from uuid import uuid4

from opentrons_shared_data.labware.dev_types import LabwareDefinition

from opentrons.api.util import (RobotBusy, robot_is_busy,
                                requires_http_protocols_disabled)
from opentrons.drivers.smoothie_drivers.driver_3_0 import SmoothieAlarm
from opentrons.broker import Broker
from opentrons.config import feature_flags as ff
from opentrons.commands.util import from_list
from opentrons.commands import types as command_types, introspection
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.context.protocol_api.protocol_context import \
    ProtocolContextImplementation
from opentrons.protocols.context.simulator.protocol_context import \
    ProtocolContextSimulation
from opentrons.protocols.parse import parse
from opentrons.protocols.types import Protocol
from opentrons.calibration_storage import helpers
from opentrons.protocol_api import (ProtocolContext,
                                    InstrumentContext,
                                    labware)
from opentrons.protocols.geometry import module_geometry
from opentrons.protocols.execution.execute import run_protocol
from opentrons.hardware_control import (API, ThreadManager,
                                        SynchronousAdapter,
                                        ExecutionCancelledError,
                                        ThreadedAsyncLock)
from opentrons.hardware_control.types import (HardwareEventType, HardwareEvent,
                                              PauseType)
from .models import Container, Instrument, Module
from .dev_types import State, StateInfo, Message, LastCommand, Error, CommandShortId

log = logging.getLogger(__name__)

VALID_STATES: Set[State] = {
    'loaded', 'running', 'finished', 'stopped', 'paused', 'error'}


class SessionManager:
    def __init__(
            self,
            hardware: SynchronousAdapter,
            loop: asyncio.AbstractEventLoop = None,
            broker: Broker = None,
            lock: ThreadedAsyncLock = None) -> None:
        self._broker = broker or Broker()
        self._loop = loop or asyncio.get_event_loop()
        self.session: Optional[Session] = None
        self._session_lock = False
        self._hardware = hardware
        self._command_logger = logging.getLogger(
            'opentrons.server.command_logger')
        self._broker.set_logger(self._command_logger)
        self._motion_lock = lock or ThreadedAsyncLock()

    @requires_http_protocols_disabled
    def create(
            self,
            name: str,
            contents: str,
            is_binary: bool = False) -> Session:
        """ Create a protocol session from either

        - a json protocol
        - a python protocol file
        - a zipped protocol bundle (deprecated, for back compat)

        No new code should be written that calls this function with
        ``is_binary=True`` and a base64'd zip in ``contents``; instead,
        use :py:meth:`.create_from_bundle`.

        :param str name: The name of the protocol
        :param str contents: The contents of the protocol; this should be
                             either the contents of the file if it can be
                             parsed directly or a base64d version of the
                             contents if it is a zip. If it is base64,
                             ``is_binary`` must be true. Do not write new
                             code that uses this.
        :param bool is_binary: ``True`` if ``contents`` is a base64'd zip.
                               Do not write new code that uses this.
        :returns Session: The created session.
        :raises Exception: If another session is simulating at the time

        .. note::

            This function is mostly (only) intended to be called via rpc.
        """
        if is_binary:
            log.warning("session.create: called with bundle")

        if self._session_lock:
            raise Exception(
                'Cannot create session while simulation in progress')

        self.clear()
        self._session_lock = True
        try:
            _contents = base64.b64decode(contents) if is_binary else contents
            session_short_id = hex(uuid4().fields[0])
            session_logger = self._command_logger.getChild(session_short_id)
            self._broker.set_logger(session_logger)
            self.session = Session.build_and_prep(
                name=name,
                contents=_contents,
                hardware=self._hardware,
                loop=self._loop,
                broker=self._broker,
                motion_lock=self._motion_lock,
                extra_labware=[])
            return self.session
        finally:
            self._session_lock = False

    @requires_http_protocols_disabled
    def create_from_bundle(self, name: str, contents: str) -> Session:
        """ Create a protocol session from a base64'd zip file.

        :param str name: The name of the protocol
        :param str contents: The contents of the zip file, base64
                             encoded
        :returns Session: The created session
        :raises Exception: If another session is simulating at the time

        .. note::

            This function is mostly (only) intended to be called via rpc.
        """
        if self._session_lock:
            raise Exception(
                'Cannot create session while simulation in progress')

        self.clear()
        self._session_lock = True
        try:
            _contents = base64.b64decode(contents)
            session_short_id = hex(uuid4().fields[0])
            session_logger = self._command_logger.getChild(session_short_id)
            self._broker.set_logger(session_logger)
            self.session = Session.build_and_prep(
                name=name,
                contents=_contents,
                hardware=self._hardware,
                loop=self._loop,
                broker=self._broker,
                motion_lock=self._motion_lock,
                extra_labware=[])
            return self.session
        finally:
            self._session_lock = False

    @requires_http_protocols_disabled
    def create_with_extra_labware(
            self,
            name: str,
            contents: str,
            extra_labware: List[LabwareDefinition]) -> Session:
        """
        Create a protocol session from a python protocol string with a list
        of extra custom labware to make available.

        :param str name: The name of the protocol
        :param str contents: The contents of the protocol file
        :param extra_labware: A list of labware definitions to make available
                              to protocol executions. This should be a list
                              of directly serialized definitions.
        :returns Session: The created session
        :raises Exception: If another session is simulating at the time

        .. note::

            This function is mostly (only) intended to be called via rpc.
        """
        if self._session_lock:
            raise Exception(
                "Cannot create session while simulation in progress")

        self.clear()
        self._session_lock = True
        try:
            session_short_id = hex(uuid4().fields[0])
            session_logger = self._command_logger.getChild(session_short_id)
            self._broker.set_logger(session_logger)
            self.session = Session.build_and_prep(
                name=name,
                contents=contents,
                hardware=self._hardware,
                loop=self._loop,
                broker=self._broker,
                motion_lock=self._motion_lock,
                extra_labware=extra_labware)
            return self.session
        finally:
            self._session_lock = False

    def clear(self) -> None:
        if self._session_lock:
            raise Exception(
                'Cannot clear session while simulation in progress')

        if self.session:
            self.session._remove_hardware_event_watcher()
            self._hardware.reset()
        self.session = None
        self._broker.set_logger(self._command_logger)

    def get_session(self) -> Optional[Session]:
        return self.session


class Session(RobotBusy):
    TOPIC: Final = 'session'

    @classmethod
    def build_and_prep(
            cls,
            name: str,
            contents: Any,
            hardware: SynchronousAdapter,
            loop: asyncio.AbstractEventLoop,
            broker: Broker,
            motion_lock: ThreadedAsyncLock,
            extra_labware: List[LabwareDefinition]
    ) -> Session:
        protocol = parse(contents, filename=name,
                         extra_labware={helpers.uri_from_definition(defn): defn
                                        for defn in extra_labware})
        sess = cls(name, protocol, hardware, loop, broker, motion_lock)
        sess.prepare()
        return sess

    def __init__(
            self, name: str, protocol: Protocol,
            hardware: SynchronousAdapter,
            loop: asyncio.AbstractEventLoop,
            broker: Broker,
            motion_lock: ThreadedAsyncLock) -> None:
        self._broker = broker
        self._default_logger = self._broker.logger
        self._sim_logger = self._broker.logger.getChild('sim')
        self._run_logger = self._broker.logger.getChild('run')
        self._loop = loop
        self.name = name
        self._protocol = protocol
        self.api_level = getattr(self._protocol, 'api_level', APIVersion(2, 0))
        self._use_v2 = self.api_level >= APIVersion(2, 0)
        log.info(
            f'Protocol API Version: {self.api_level}; '
            f'Protocol kind: {type(self._protocol)}')

        # self.metadata is exposed via rpc
        self.metadata = getattr(self._protocol, 'metadata', {})

        self._hardware = hardware
        self._simulating_ctx = ProtocolContext.build_using(
            implementation=ProtocolContextImplementation.build_using(
                self._protocol),
            protocol=self._protocol,
            loop=self._loop,
            broker=self._broker
        )

        self.state: 'State' = None
        #: The current state
        self.stateInfo: 'StateInfo' = {}
        #: A message associated with the current state
        self.commands: List[command_types.CommandMessage] = []
        self.command_log: Dict[int, int] = {}
        self.errors: List[Error] = []

        # Underlying objects harvested from commands, internal
        self._containers: List[labware.Labware] = []
        self._instruments: List[InstrumentContext] = []
        self._modules: List[module_geometry.ModuleGeometry] = []
        self._interactions: List[Tuple[InstrumentContext, labware.Labware]] = []

        # RPC-safe models of objects harvested from commands
        self.instruments: Optional[List[Instrument]] = None
        self.containers: Optional[List[Container]] = None
        self.modules: Optional[List[Module]] = None
        self.protocol_text = protocol.text

        self.startTime: Optional[float] = None
        self._motion_lock = motion_lock
        self._event_watcher = None
        self.door_state: Optional[str] = None
        self.blocked: Optional[bool] = None

    @property
    def busy_lock(self) -> ThreadedAsyncLock:
        return self._motion_lock

    def _hw_iface(self) -> SynchronousAdapter:
        return self._hardware

    def prepare(self) -> None:
        self.refresh()

    def get_instruments(self) -> List[Instrument]:
        return [
            Instrument(
                instrument=instrument,
                containers=[
                    container
                    for _instrument, container in
                    self._interactions
                    if _instrument == instrument
                ],
                context=self._use_v2 and self._simulating_ctx)
            for instrument in self._instruments
        ]

    def get_containers(self) -> List[Container]:
        return [
            Container(
                container=container,
                instruments=[
                    instrument
                    for instrument, _container in
                    self._interactions
                    if _container == container
                ],
                context=self._use_v2 and self._simulating_ctx)
            for container in self._containers
        ]

    def get_modules(self) -> List[Module]:
        return [
            Module(module=module,
                   context=self._use_v2 and self._simulating_ctx)
            for module in self._modules
        ]

    def clear_logs(self) -> None:
        self.command_log.clear()
        self.errors.clear()

    @robot_is_busy
    def _simulate(self) -> List[CommandShortId]:
        self._reset()

        stack: List[command_types.CommandMessage] = []
        res: List[CommandShortId] = []
        commands: List[command_types.CommandPayload] = []

        self._containers.clear()
        self._instruments.clear()
        self._modules.clear()
        self._interactions.clear()

        def on_command(message: command_types.CommandMessage) -> None:
            payload = message['payload']
            description = payload.get('text', '').format(
                **payload
            )

            if message['$'] == 'before':
                level = len(stack)

                stack.append(message)
                commands.append(payload)

                res.append(
                    {
                        'level': level,
                        'description': description,
                        'id': len(res)})
            else:
                stack.pop()
        unsubscribe = self._broker.subscribe(command_types.COMMAND, on_command)

        try:
            # ensure actual pipettes are cached before driver is disconnected
            self._hardware.cache_instruments()
            instrs = {}
            for mount, pip in self._hardware.attached_instruments.items():
                if pip:
                    instrs[mount] = {'model': pip['model'],
                                     'id': pip.get('pipette_id', '')}
            sync_sim = ThreadManager(
                API.build_hardware_simulator,
                instrs,
                [mod.name()
                 for mod in self._hardware.attached_modules],
                strict_attached_instruments=False
            ).sync
            sync_sim.home()

            # Use feature flag to determine ProtocolContext implementation
            # class: real or simulating.
            impl_class = ProtocolContextImplementation \
                if not ff.enable_fast_protocol_upload() \
                else ProtocolContextSimulation

            ctx_impl = impl_class.build_using(
                self._protocol,
                hardware=sync_sim)
            self._simulating_ctx = ProtocolContext.build_using(
                protocol=self._protocol,
                loop=self._loop,
                broker=self._broker,
                implementation=ctx_impl
            )
            run_protocol(self._protocol,
                         context=self._simulating_ctx)

        finally:
            unsubscribe()

            instruments, containers, modules, interactions = _accumulate(
                [_get_labware(command) for command in commands])

            self._containers.extend(_dedupe(containers))
            self._instruments.extend(_dedupe(
                instruments
                + list(self._simulating_ctx.loaded_instruments.values())))
            self._modules.extend(_dedupe(
                modules
                + [m._geometry
                   for m in self._simulating_ctx.loaded_modules.values()]))
            self._interactions.extend(_dedupe(interactions))

            # Labware calibration happens after simulation and before run, so
            # we have to clear the tips if they are left on after simulation
            # to ensure that the instruments are in the expected state at the
            # beginning of the labware calibration flow

        return res

    def refresh(self) -> None:
        self._reset()

        try:
            self._broker.set_logger(self._sim_logger)
            commands = self._simulate()
        except Exception:
            raise
        finally:
            self._broker.set_logger(self._default_logger)

        self.commands = from_list(commands)

        self.containers = self.get_containers()
        self.instruments = self.get_instruments()
        self.modules = self.get_modules()
        self.startTime = None
        self.set_state('loaded')

    def stop(self) -> None:
        self._hw_iface().halt()
        with self._motion_lock.lock():
            try:
                self._hw_iface().stop()
            except asyncio.CancelledError:
                pass
        self.set_state('stopped')

    def pause(self,
              reason: str = None,
              user_message: str = None,
              duration: float = None) -> None:
        self._hardware.pause(PauseType.PAUSE)
        self.set_state(
            'paused', reason=reason,
            user_message=user_message, duration=duration)

    def resume(self) -> None:
        if not self.blocked:
            self._hardware.resume(PauseType.PAUSE)
            self.set_state('running')

    def _start_hardware_event_watcher(self) -> None:
        if not callable(self._event_watcher):
            # initialize and update window switch state
            self.door_state = str(self._hardware.door_state)
            self.blocked = self._hardware._pause_manager.blocked_by_door
            log.info('Starting hardware event watcher')
            self._event_watcher = self._hardware.register_callback(
                self._handle_hardware_event)
        else:
            log.warning("Cannot start new hardware event watcher "
                        "when one already exists")

    def _remove_hardware_event_watcher(self) -> None:
        if self._event_watcher and callable(self._event_watcher):
            self._event_watcher()
            self._event_watcher = None

    def _handle_hardware_event(self, hw_event: HardwareEvent) -> None:
        if hw_event.event == HardwareEventType.DOOR_SWITCH_CHANGE:
            self.door_state = str(hw_event.new_state)
            self.blocked = hw_event.blocking
            if self.blocked and self.state == 'running':
                self.pause('Robot door is open')
            else:
                self._on_state_changed()

    @robot_is_busy
    def _run(self) -> None:
        def on_command(message: command_types.CommandMessage) -> None:
            if message['$'] == 'before':
                self.log_append()
            if message['name'] == command_types.PAUSE:
                self.set_state('paused',
                               reason='The protocol paused execution',
                               user_message=message['payload']['userMessage'])
            if message['name'] == command_types.RESUME:
                self.set_state('running')

        self._reset()

        _unsubscribe = self._broker.subscribe(
            command_types.COMMAND, on_command)

        self.startTime = now()
        self.set_state('running')

        try:
            self.resume()
            self._pre_run_hooks()
            self._hardware.cache_instruments()
            self._hardware.reset_instrument()
            ctx_impl = ProtocolContextImplementation.build_using(
                self._protocol,
                extra_labware=getattr(self._protocol, 'extra_labware', {}))
            ctx = ProtocolContext.build_using(
                protocol=self._protocol,
                implementation=ctx_impl,
                loop=self._loop,
                broker=self._broker
            )
            ctx.connect(self._hardware)
            ctx.home()
            run_protocol(self._protocol, context=ctx)

            # If the last command in a protocol was a pause, the protocol
            # will immediately finish executing because there's no smoothie
            # command to block... except the home that's about to happen,
            # which will confuse the app and lock it up. So we need to
            # do our own pause here, and sleep the thread until/unless the
            # app resumes us.
            #
            # Cancelling from the app during this pause will result in the
            # smoothie giving us an error during the subsequent home, which
            # is tragic but expected.
            while self.state == 'paused':
                sleep(0.1)
            self.set_state('finished')
            self._hw_iface().home()
        except (SmoothieAlarm, asyncio.CancelledError,
                ExecutionCancelledError):
            log.info("Protocol cancelled")
        except Exception as e:
            log.exception("Exception during run:")
            self.error_append(e)
            self.set_state('error')
            raise e
        finally:
            _unsubscribe()

    def run(self) -> None:
        if not self.blocked:
            try:
                self._broker.set_logger(self._run_logger)
                self._run()
            except Exception:
                raise
            finally:
                self._broker.set_logger(self._default_logger)
        else:
            raise RuntimeError(
                'Protocol is blocked and cannot run. Make sure robot door '
                'is closed before running.')

    def set_state(self, state: 'State',
                  reason: str = None,
                  user_message: str = None,
                  duration: float = None) -> None:
        log.info(f"Session.set_state state={state} "
                 f"user_message={user_message} duration={duration} "
                 f"reason={reason}")
        if state not in VALID_STATES:
            raise ValueError(
                'Invalid state: {0}. Valid states are: {1}'
                .format(state, VALID_STATES))
        self.state = state
        if user_message:
            self.stateInfo['userMessage'] = user_message
        else:
            self.stateInfo.pop('userMessage', None)
        if reason:
            self.stateInfo['message'] = reason
        else:
            self.stateInfo.pop('message', None)
        if duration:
            self.stateInfo['estimatedDuration'] = duration
        else:
            self.stateInfo.pop('estimatedDuration', None)
        if self.startTime:
            self.stateInfo['changedAt'] = now() - self.startTime
        else:
            self.stateInfo.pop('changedAt', None)
        self._on_state_changed()

    def log_append(self) -> None:
        self.command_log.update({
            len(self.command_log): now()})
        self._on_state_changed()

    def error_append(self, error: Exception) -> None:
        self.errors.append(
            {
                'timestamp': now(),
                'error': error
            }
        )

    def _reset(self) -> None:
        self._hw_iface().reset()
        self.clear_logs()
        # unregister existing event watcher
        self._remove_hardware_event_watcher()
        self._start_hardware_event_watcher()

    def _snapshot(self) -> Message:
        if self.state == 'loaded':
            payload: Any = copy(self)
        else:
            if self.command_log.keys():
                idx = sorted(self.command_log.keys())[-1]
                timestamp = self.command_log[idx]
                last_command: Optional[LastCommand]\
                    = {'id': idx, 'handledAt': timestamp}
            else:
                last_command = None

            payload = {
                'state': self.state,
                'stateInfo': self.stateInfo,
                'startTime': self.startTime,
                'doorState': self.door_state,
                'blocked': self.blocked,
                'errors': self.errors,
                'lastCommand': last_command
            }
        return {
            'topic': Session.TOPIC,
            'payload': payload
        }

    def _on_state_changed(self) -> None:
        snap = self._snapshot()
        self._broker.publish(Session.TOPIC, snap)

    def _pre_run_hooks(self) -> None:
        self._hw_iface().home_z()


CommandReferents = Tuple[
    List[InstrumentContext], List[labware.Labware],
    List[module_geometry.ModuleGeometry],
    List[Tuple[InstrumentContext, labware.Labware]]]


def _accumulate(iterable: Sequence[CommandReferents]) -> CommandReferents:

    def _reducer(acc: CommandReferents, item: CommandReferents) -> CommandReferents:
        return (acc[0] + item[0],
                acc[1] + item[1],
                acc[2] + item[2],
                acc[3] + item[3])

    return reduce(
        _reducer,
        iterable,
        cast(CommandReferents, ([], [], [], [])))


It = TypeVar('It')


def _dedupe(iterable: Sequence[It]) -> Iterator[It]:
    acc: Set[It] = set()

    for item in iterable:
        if item not in acc:
            acc.add(item)
            yield item


def now() -> int:
    return int(time() * 1000)


def _get_labware(
        command: command_types.CommandPayload) -> CommandReferents:
    interactions: List[Tuple[InstrumentContext, labware.Labware]] = []

    try:
        instruments, containers, modules = introspection.get_referred_objects(command)
    except ValueError:
        log.exception(f'Cant handle location in command {command!r}')
        instruments = []
        containers = []
        modules = []

    for instrument in instruments:
        interactions.extend([(instrument, container) for container in containers])

    return instruments, containers, modules, interactions



import logging
from typing import Dict, Optional, Set, List, TYPE_CHECKING
from collections import OrderedDict

from opentrons import types, API
from opentrons.protocols.api_support.types import APIVersion
from opentrons.config import feature_flags as fflags
from opentrons.hardware_control.types import DoorState, PauseType
from opentrons.hardware_control import SynchronousAdapter
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.geometry.deck import Deck
from opentrons.protocols.geometry import module_geometry
from opentrons.protocols.geometry.deck_item import DeckItem
from opentrons.protocols.context.protocol_api.instrument_context import \
    InstrumentContextImplementation
from opentrons.protocols.context.instrument import \
    AbstractInstrument
from opentrons.protocols.context.labware import \
    AbstractLabware
from opentrons.protocols.context.protocol import \
    AbstractProtocol, InstrumentDict, LoadModuleResult
from opentrons.protocols.api_support.util import (
    AxisMaxSpeeds, HardwareToManage, HardwareManager)
from opentrons.protocols.labware import load_from_definition, \
    get_labware_definition
from opentrons.protocols.types import Protocol
from opentrons_shared_data.labware.dev_types import LabwareDefinition

if TYPE_CHECKING:
    from opentrons.hardware_control.modules import AbstractModule


MODULE_LOG = logging.getLogger(__name__)

SHORT_TRASH_DECK = 'ot2_short_trash'
STANDARD_DECK = 'ot2_standard'


class ProtocolContextImplementation(AbstractProtocol):

    def __init__(self,
                 api_version: Optional[APIVersion] = None,
                 hardware: Optional[HardwareToManage] = None,
                 bundled_labware: Dict[str, LabwareDefinition] = None,
                 bundled_data: Dict[str, bytes] = None,
                 extra_labware: Dict[str, LabwareDefinition] = None,
                 ) -> None:
        """ Build a :py:class:`.ProtocolContextImplementation`.

        :param api_version: The API version to use. If this is ``None``, uses
                            the max supported version.
        :param hardware: An optional hardware controller to link to. If not
                         specified, a new simulator will be created.
        :param bundled_labware: A dict mapping labware URIs to definitions.
                                This is used when executing bundled protocols,
                                and if specified will be the only allowed
                                source for labware definitions, excluding the
                                built in definitions and anything in
                                ``extra_labware``.
        :param bundled_data: A dict mapping filenames to the contents of data
                             files. Can be used by the protocol, since it is
                             exposed as
                             :py:attr:`.ProtocolContext.bundled_data`
        :param extra_labware: A dict mapping labware URIs to definitions. These
                              URIs are searched during :py:meth:`.load_labware`
                              in addition to the system definitions (if
                              ``bundled_labware`` was not specified). Used to
                              provide custom labware definitions.
        """
        # TODO AL 20201110 - Find a way to omit api_version from this class.
        #  it is used in creating module geometry and instrument context.
        self._api_version = api_version or MAX_SUPPORTED_VERSION
        deck_load_name = SHORT_TRASH_DECK if fflags.short_fixed_trash() \
            else STANDARD_DECK
        self._deck_layout = Deck(load_name=deck_load_name)
        self._instruments: InstrumentDict = {
            mount: None for mount in types.Mount
        }
        self._modules: List[LoadModuleResult] = []

        self._hw_manager = HardwareManager(hardware)
        self._log = MODULE_LOG.getChild(self.__class__.__name__)

        self._bundled_labware = bundled_labware
        self._extra_labware = extra_labware or {}

        self._bundled_data: Dict[str, bytes] = bundled_data or {}
        self._default_max_speeds = AxisMaxSpeeds()
        self._last_location: Optional[types.Location] = None
        self._last_mount: Optional[types.Mount] = None
        self._loaded_modules: Set['AbstractModule'] = set()

    @classmethod
    def build_using(cls,
                    protocol: Protocol,
                    *args, **kwargs):
        """ Build an API instance for the specified parsed protocol

        This is used internally to provision the context with bundle
        contents or api levels.
        """
        kwargs['bundled_data'] = getattr(protocol, 'bundled_data', None)
        kwargs['bundled_labware'] = getattr(protocol, 'bundled_labware', None)
        kwargs['extra_labware'] = getattr(protocol, 'extra_labware', None)
        kwargs['api_version'] = getattr(
            protocol, 'api_level', MAX_SUPPORTED_VERSION)
        return cls(*args, **kwargs)

    def get_bundled_data(self) -> Dict[str, bytes]:
        """Extra bundled data."""
        # TODO AL 20201110 - This should be removed along with the bundling
        #  feature as we move to HTTP based protocol execution.
        return self._bundled_data

    def get_bundled_labware(self) -> Optional[Dict[str, LabwareDefinition]]:
        """Bundled labware defintion."""
        # TODO AL 20201110 - This should be removed along with the bundling
        #  feature as we move to HTTP based protocol execution.
        return self._bundled_labware

    def get_extra_labware(self) -> Optional[Dict[str, LabwareDefinition]]:
        """Extra labware definitions."""
        return self._extra_labware

    def cleanup(self) -> None:
        """Protocol context clean up."""
        pass

    def get_max_speeds(self) -> AxisMaxSpeeds:
        """Get the maximum axis speeds."""
        return self._default_max_speeds

    def get_hardware(self) -> HardwareManager:
        """Access to the hardware manager."""
        return self._hw_manager

    def connect(self, hardware: API) -> None:
        """Connect to the hardware."""
        self._hw_manager.set_hw(hardware)
        self._hw_manager.hardware.cache_instruments()

    def disconnect(self) -> None:
        """"Disconnect from the hardware."""
        self._hw_manager.reset_hw()

    def is_simulating(self) -> bool:
        """Returns true if hardware is being simulated."""
        return self._hw_manager.hardware.is_simulator

    def load_labware_from_definition(
            self,
            labware_def: LabwareDefinition,
            location: types.DeckLocation,
            label: Optional[str]) -> AbstractLabware:
        """Load a labware from definition"""
        parent = self.get_deck().position_for(location)
        labware_obj = load_from_definition(labware_def, parent, label)
        self._deck_layout[location] = labware_obj
        return labware_obj

    def load_labware(
            self,
            load_name: str,
            location: types.DeckLocation,
            label: Optional[str],
            namespace: Optional[str],
            version: Optional[int]) -> AbstractLabware:
        """Load a labware."""
        labware_def = get_labware_definition(
            load_name, namespace, version,
            bundled_defs=self._bundled_labware,
            extra_defs=self._extra_labware)
        return self.load_labware_from_definition(
            labware_def, location, label)

    def load_module(
            self,
            module_name: str,
            location: Optional[types.DeckLocation],
            configuration: Optional[str]) -> Optional[LoadModuleResult]:
        """Load a module."""
        resolved_model = module_geometry.resolve_module_model(module_name)
        resolved_type = module_geometry.resolve_module_type(resolved_model)
        resolved_location = self._deck_layout.resolve_module_location(
            resolved_type, location)

        # Load the geometry
        geometry = module_geometry.load_module(
            model=resolved_model,
            parent=self._deck_layout.position_for(resolved_location),
            api_level=self._api_version,
            configuration=configuration)

        # Try to find in the hardware instance
        available_modules, simulating_module = self._hw_manager.hardware.find_modules(
            resolved_model, resolved_type)

        hc_mod_instance = None
        for mod in available_modules:
            compatible = module_geometry.models_compatible(
                module_geometry.module_model_from_string(mod.model()),
                resolved_model)
            if compatible and mod not in self._loaded_modules:
                self._loaded_modules.add(mod)
                hc_mod_instance = SynchronousAdapter(mod)
                break

        if simulating_module and not hc_mod_instance:
            hc_mod_instance = SynchronousAdapter(simulating_module)

        if not hc_mod_instance:
            return None

        result = LoadModuleResult(type=resolved_type,
                                  geometry=geometry,
                                  module=hc_mod_instance)

        self._modules.append(result)
        self._deck_layout[resolved_location] = geometry
        return result

    def get_loaded_modules(self) -> Dict[int, LoadModuleResult]:
        """Get a mapping of deck location to loaded module."""
        return OrderedDict({int(str(module.geometry.parent)): module
                            for module in self._modules})

    def load_instrument(self,
                        instrument_name: str,
                        mount: types.Mount,
                        replace: bool) -> AbstractInstrument:
        """Load an instrument."""
        instr = self._instruments[mount]
        if instr and not replace:
            raise RuntimeError(
                f"Instrument already present in {mount.name.lower()} "
                f"mount: {instr.get_instrument_name()}")

        attached = {att_mount: instr.get('name', None)
                    for att_mount, instr
                    in self._hw_manager.hardware.attached_instruments.items()
                    if instr}
        attached[mount] = instrument_name
        self._hw_manager.hardware.cache_instruments(attached)
        # If the cache call didn’t raise, the instrument is attached
        new_instr = InstrumentContextImplementation(
            api_version=self._api_version,
            protocol_interface=self,
            mount=mount,
            instrument_name=instrument_name,
            default_speed=400.0
        )
        self._instruments[mount] = new_instr
        self._log.info("Instrument {} loaded".format(new_instr))
        return new_instr

    def get_loaded_instruments(self) -> InstrumentDict:
        """Get a mapping of mount to instrument."""
        return self._instruments

    def pause(self, msg: Optional[str]) -> None:
        """Pause the protocol."""
        self._hw_manager.hardware.pause(PauseType.PAUSE)

    def resume(self) -> None:
        """Result the protocol."""
        self._hw_manager.hardware.resume(PauseType.PAUSE)

    def comment(self, msg: str) -> None:
        """Add comment to run log."""
        pass

    def delay(self, seconds: float, msg: Optional[str]) -> None:
        """Delay execution for x seconds."""
        self._hw_manager.hardware.delay(seconds)

    def home(self) -> None:
        """Home the robot."""
        self.set_last_location(None)
        self._hw_manager.hardware.home()

    def get_deck(self) -> Deck:
        """Get the deck layout."""
        return self._deck_layout

    def get_fixed_trash(self) -> DeckItem:
        """The trash fixed to slot 12 of the robot deck."""
        trash = self._deck_layout['12']
        if not trash:
            raise RuntimeError("Robot must have a trash container in 12")
        return trash

    def set_rail_lights(self, on: bool) -> None:
        """Set the rail light state."""
        self._hw_manager.hardware.set_lights(rails=on)

    def get_rail_lights_on(self) -> bool:
        """Get the rail light state."""
        return self._hw_manager.hardware.get_lights()['rails']

    def door_closed(self) -> bool:
        """Check if door is closed."""
        return DoorState.CLOSED == self._hw_manager.hardware.door_state

    def get_last_location(
        self,
        mount: Optional[types.Mount] = None,
    ) -> Optional[types.Location]:
        """Get the most recent moved to location."""
        if mount is None or mount == self._last_mount:
            return self._last_location

        return None

    def set_last_location(
        self,
        location: Optional[types.Location],
        mount: Optional[types.Mount] = None,
    ) -> None:
        """Set the most recent moved to location."""
        self._last_location = location
        self._last_mount = mount



from typing import List, Dict, Optional

from opentrons.calibration_storage import helpers
from opentrons.protocols.geometry.labware_geometry import LabwareGeometry
from opentrons.protocols.geometry.well_geometry import WellGeometry
from opentrons.protocols.context.labware import \
    AbstractLabware
from opentrons.protocols.api_support.tip_tracker import TipTracker
from opentrons.protocols.context.well import WellImplementation
from opentrons.protocols.api_support.well_grid import WellGrid
from opentrons.types import Point, Location
from opentrons_shared_data.labware.dev_types import LabwareParameters, \
    LabwareDefinition


class LabwareImplementation(AbstractLabware):

    def __init__(self,
                 definition: LabwareDefinition,
                 parent: Location,
                 label: Optional[str] = None):
        """
        Construct an implementation of a labware object.

        :param definition: A dict representing all required data for a labware,
                           including metadata such as the display name of the
                           labware, a definition of the order to iterate over
                           wells, the shape of wells (shape, physical
                           dimensions, etc), and so on. The correct shape of
                           this definition is handled by the "labware-designer"
                           project in the Opentrons/opentrons repo.
        :param parent: A :py:class:`.Location` representing the location where
                       the front and left most point of the outside of the
                       labware is (often the front-left corner of a slot on the
                       deck).
        :param label: An optional label to use instead of the displayName
                      from the definition's metadata element
        """
        if label:
            dn = label
            self._name = dn
        else:
            dn = definition['metadata']['displayName']
            self._name = definition['parameters']['loadName']

        self._display_name = f"{dn} on {str(parent.labware)}"
        # Directly from definition
        self._well_definition = definition['wells']
        self._parameters = definition['parameters']
        self._definition = definition

        self._geometry = LabwareGeometry(definition, parent)
        # flatten list of list of well names.
        self._ordering = [
            well for col in definition['ordering'] for well in col
        ]
        self._wells: List[WellImplementation] = []
        self._well_name_grid = WellGrid(wells=self._wells)
        self._tip_tracker = TipTracker(
            columns=self._well_name_grid.get_columns()
        )

        self._calibrated_offset = Point(0, 0, 0)
        # Will cause building of the wells
        self.set_calibration(self._calibrated_offset)

    def get_uri(self) -> str:
        return helpers.uri_from_definition(self._definition)

    def get_display_name(self) -> str:
        return self._display_name

    def get_name(self) -> str:
        return self._name

    def set_name(self, new_name: str) -> None:
        self._name = new_name

    def get_definition(self) -> LabwareDefinition:
        return self._definition

    def get_parameters(self) -> LabwareParameters:
        return self._parameters

    def get_quirks(self) -> List[str]:
        return self._parameters.get('quirks', [])

    def set_calibration(self, delta: Point) -> None:
        self._calibrated_offset = Point(
            x=self._geometry.offset.x + delta.x,
            y=self._geometry.offset.y + delta.y,
            z=self._geometry.offset.z + delta.z
        )
        # The wells must be rebuilt
        self._wells = self._build_wells()
        self._well_name_grid = WellGrid(wells=self._wells)
        self._tip_tracker = TipTracker(
            columns=self._well_name_grid.get_columns()
        )

    def get_calibrated_offset(self) -> Point:
        return self._calibrated_offset

    def is_tiprack(self) -> bool:
        return self._parameters['isTiprack']

    def get_tip_length(self) -> float:
        return self._parameters['tipLength']

    def set_tip_length(self, length: float):
        self._parameters['tipLength'] = length

    def reset_tips(self) -> None:
        if self.is_tiprack():
            for well in self._wells:
                well.set_has_tip(True)

    def get_tip_tracker(self) -> TipTracker:
        return self._tip_tracker

    def get_well_grid(self) -> WellGrid:
        return self._well_name_grid

    def get_wells(self) -> List[WellImplementation]:
        return self._wells

    def get_wells_by_name(self) -> Dict[str, WellImplementation]:
        return {
            well.get_name(): well for well in self._wells
        }

    def get_geometry(self) -> LabwareGeometry:
        return self._geometry

    @property
    def highest_z(self):
        return self._geometry.z_dimension + self._calibrated_offset.z

    @property
    def separate_calibration(self) -> bool:
        return False

    @property
    def load_name(self) -> str:
        return self._parameters['loadName']

    def _build_wells(self) -> List[WellImplementation]:
        return [
            WellImplementation(
                well_geometry=WellGeometry(
                    well_props=self._well_definition[well],
                    parent_point=self._calibrated_offset,
                    parent_object=self
                ),
                display_name="{} of {}".format(well, self._display_name),
                has_tip=self.is_tiprack(),
                name=well
            )
            for well in self._ordering
        ]



import logging
from logging.config import dictConfig
import sys
from typing import Any, Dict

from opentrons.config import CONFIG, ARCHITECTURE, SystemArchitecture


def _balena_config(level_value: int) -> Dict[str, Any]:
    serial_log_filename = CONFIG['serial_log_file']
    api_log_filename = CONFIG['api_log_file']
    return {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'basic': {
                'format':
                '%(asctime)s %(name)s %(levelname)s [Line %(lineno)s] %(message)s'
            },
        },
        'handlers': {
            'debug': {
                'class': 'logging.StreamHandler',
                'formatter': 'basic',
                'level': level_value
            },
            'serial': {
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'basic',
                'filename': serial_log_filename,
                'maxBytes': 5000000,
                'level': logging.DEBUG,
                'backupCount': 3
            },
            'api': {
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'basic',
                'filename': api_log_filename,
                'maxBytes': 1000000,
                'level': logging.DEBUG,
                'backupCount': 5
            }
        },
        'loggers': {
            'opentrons': {
                'handlers': ['debug', 'api'],
                'level': level_value,
            },
            'opentrons.deck_calibration': {
                'handlers': ['debug', 'api'],
                'level': level_value,
            },
            'opentrons.drivers.serial_communication': {
                'handlers': ['serial'],
                'level': logging.DEBUG,
                'propagate': False
            },
            '__main__': {
                'handlers': ['api'],
                'level': level_value
            },
        }
    }


def _buildroot_config(level_value: int) -> Dict[str, Any]:
    # Import systemd.journald here since it is generally unavailble on non
    # linux systems and we probably don't want to use it on linux desktops
    # either
    return {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'message_only': {
                'format': '%(message)s'
            },
        },
        'handlers': {
            'api': {
                'class': 'systemd.journal.JournalHandler',
                'level': logging.DEBUG,
                'formatter': 'message_only',
                'SYSLOG_IDENTIFIER': 'opentrons-api',
            },
            'serial': {
                'class': 'systemd.journal.JournalHandler',
                'level': logging.DEBUG,
                'formatter': 'message_only',
                'SYSLOG_IDENTIFIER': 'opentrons-api-serial',
            }
        },
        'loggers': {
            'opentrons.drivers.serial_communication': {
                'handlers': ['serial'],
                'level': logging.DEBUG,
                'propagate': False
            },
            'opentrons': {
                'handlers': ['api'],
                'level': level_value,
            },
            '__main__': {
                'handlers': ['api'],
                'level': level_value
            },
        },

    }


def _host_config(level_value: int) -> Dict[str, Any]:
    """ Host logging, for now the same as balena logging
    """
    return _balena_config(level_value)


def _config(arch: SystemArchitecture, level_value: int) -> Dict[str, Any]:
    return {
        SystemArchitecture.BALENA: _balena_config,
        SystemArchitecture.BUILDROOT: _buildroot_config,
        SystemArchitecture.HOST: _host_config
    }[arch](level_value)


def log_init(level_name: str):
    """
    Function that sets log levels and format strings. Checks for the
    OT_API_LOG_LEVEL environment variable otherwise defaults to INFO
    """
    fallback_log_level = 'INFO'
    ot_log_level = level_name.upper()
    if ot_log_level not in logging._nameToLevel:
        sys.stderr.write(
            f'OT Log Level {ot_log_level} not found. '
            f'Defaulting to {fallback_log_level}\n')
        ot_log_level = fallback_log_level
    level_value = logging._nameToLevel[ot_log_level]
    logging_config = _config(ARCHITECTURE, level_value)
    dictConfig(logging_config)



from __future__ import annotations
""" Classes and functions for pipette state tracking
"""
from dataclasses import asdict, replace
import logging
from typing import Any, Dict, Optional, Set, Tuple, Union, TYPE_CHECKING

from opentrons.types import Point
from opentrons.calibration_storage.types import PipetteOffsetByPipetteMount
from opentrons.config import pipette_config, robot_configs
from opentrons.config.types import RobotConfig
from opentrons.drivers.types import MoveSplit
from .types import CriticalPoint, BoardRevision


if TYPE_CHECKING:
    from opentrons_shared_data.pipette.dev_types import (
        UlPerMmAction, PipetteName, PipetteModel
    )
    from .dev_types import InstrumentHardwareConfigs


RECONFIG_KEYS = {'quirks'}


mod_log = logging.getLogger(__name__)


class Pipette:
    """ A class to gather and track pipette state and configs.

    This class should not touch hardware or call back out to the hardware
    control API. Its only purpose is to gather state.
    """

    DictType = Dict[str, Union[str, float, bool]]
    #: The type of this data class as a dict

    def __init__(
            self,
            config: pipette_config.PipetteConfig,
            pipette_offset_cal: PipetteOffsetByPipetteMount,
            pipette_id: str = None) -> None:
        self._config = config
        self._pipette_offset = pipette_offset_cal
        self._acting_as = self._config.name
        self._name = self._config.name
        self._model = self._config.model
        self._nozzle_offset = self._config.nozzle_offset
        self._current_volume = 0.0
        self._working_volume = self._config.max_volume
        self._current_tip_length = 0.0
        self._current_tiprack_diameter = 0.0
        self._fallback_tip_length = self._config.tip_length
        self._tip_overlap_map = self._config.tip_overlap
        self._has_tip = False
        self._pipette_id = pipette_id
        self._log = mod_log.getChild(self._pipette_id
                                     if self._pipette_id else '<unknown>')
        self._log.info("loaded: {}, pipette offset: {}".format(
            config.model, self._pipette_offset.offset))
        self.ready_to_aspirate = False
        #: True if ready to aspirate
        self._aspirate_flow_rate\
            = self._config.default_aspirate_flow_rates['2.0']
        self._dispense_flow_rate\
            = self._config.default_dispense_flow_rates['2.0']
        self._blow_out_flow_rate\
            = self._config.default_blow_out_flow_rates['2.0']
        # cache a dict representation of config for improved performance of
        # as_dict.
        self._config_as_dict = asdict(config)

    def act_as(self, name: PipetteName):
        """ Reconfigure to act as ``name``. ``name`` must be either the
        actual name of the pipette, or a name in its back-compatibility
        config.
        """
        if name == self._acting_as:
            return

        assert name in self._config.back_compat_names + [self.name],\
            f'{self._name} is not back-compatible with {name}'
        name_conf = pipette_config.name_config()
        bc_conf = name_conf[name]
        self.working_volume = bc_conf['maxVolume']
        self.update_config_item('min_volume', bc_conf['minVolume'])
        self.update_config_item('max_volume', bc_conf['maxVolume'])

    @property
    def acting_as(self) -> PipetteName:
        return self._acting_as

    def update_pipette_offset(self, offset_cal: PipetteOffsetByPipetteMount):
        self._log.info("updating pipette offset to {}"
                       .format(offset_cal.offset))
        self._pipette_offset = offset_cal

    @property
    def config(self) -> pipette_config.PipetteConfig:
        return self._config

    @property
    def nozzle_offset(self) -> Tuple[float, float, float]:
        return self._nozzle_offset

    def update_config_item(self, elem_name: str, elem_val: Any):
        self._log.info("updated config: {}={}".format(elem_name, elem_val))
        self._config = replace(self._config,
                               **{elem_name: elem_val})
        # Update the cached dict representation
        self._config_as_dict = asdict(self._config)

    @property
    def name(self) -> PipetteName:
        return self._name

    @property
    def model(self) -> PipetteModel:
        return self._model

    @property
    def pipette_id(self) -> Optional[str]:
        return self._pipette_id

    def critical_point(self, cp_override: CriticalPoint = None) -> Point:
        """
        The vector from the pipette's origin to its critical point. The
        critical point for a pipette is the end of the nozzle if no tip is
        attached, or the end of the tip if a tip is attached.

        If `cp_override` is specified and valid - so is either
        :py:attr:`CriticalPoint.NOZZLE` or :py:attr:`CriticalPoint.TIP` when
        we have a tip, or :py:attr:`CriticalPoint.XY_CENTER` - the specified
        critical point will be used.
        """
        instr = Point(*self._pipette_offset.offset)
        offsets = self.nozzle_offset

        if not self.has_tip or cp_override == CriticalPoint.NOZZLE:
            cp_type = CriticalPoint.NOZZLE
            tip_length = 0.0
        else:
            cp_type = CriticalPoint.TIP
            tip_length = self.current_tip_length
        if cp_override == CriticalPoint.XY_CENTER:
            mod_offset_xy = [0, 0, offsets[2]]
            cp_type = CriticalPoint.XY_CENTER
        elif cp_override == CriticalPoint.FRONT_NOZZLE:
            mod_offset_xy = [
                0, -offsets[1], offsets[2]]
            cp_type = CriticalPoint.FRONT_NOZZLE
        else:
            mod_offset_xy = list(offsets)
        mod_and_tip = Point(mod_offset_xy[0],
                            mod_offset_xy[1],
                            mod_offset_xy[2] - tip_length)

        cp = mod_and_tip + instr

        if self._log.isEnabledFor(logging.DEBUG):
            info_str = 'cp: {}{}: {} (from: '\
                .format(cp_type,
                        ' (from override)' if cp_override else '',
                        cp)
            info_str += 'model offset: {} + instrument offset: {}'\
                .format(mod_offset_xy, instr)
            info_str += ' - tip_length: {}'.format(tip_length)
            info_str += ')'
            self._log.debug(info_str)

        return cp

    @property
    def current_volume(self) -> float:
        """ The amount of liquid currently aspirated """
        return self._current_volume

    @property
    def current_tip_length(self) -> float:
        """ The length of the current tip attached (0.0 if no tip) """
        return self._current_tip_length

    @current_tip_length.setter
    def current_tip_length(self, tip_length: float):
        self._current_tip_length = tip_length

    @property
    def current_tiprack_diameter(self) -> float:
        """ The diameter of the current tip rack (0.0 if no tip) """
        return self._current_tiprack_diameter

    @current_tiprack_diameter.setter
    def current_tiprack_diameter(self, diameter: float):
        self._current_tiprack_diameter = diameter

    @property
    def aspirate_flow_rate(self) -> float:
        """ Current active flow rate (not config value)"""
        return self._aspirate_flow_rate

    @aspirate_flow_rate.setter
    def aspirate_flow_rate(self, new_flow_rate: float):
        assert new_flow_rate > 0
        self._aspirate_flow_rate = new_flow_rate

    @property
    def dispense_flow_rate(self) -> float:
        """ Current active flow rate (not config value)"""
        return self._dispense_flow_rate

    @dispense_flow_rate.setter
    def dispense_flow_rate(self, new_flow_rate: float):
        assert new_flow_rate > 0
        self._dispense_flow_rate = new_flow_rate

    @property
    def blow_out_flow_rate(self) -> float:
        """ Current active flow rate (not config value)"""
        return self._blow_out_flow_rate

    @blow_out_flow_rate.setter
    def blow_out_flow_rate(self, new_flow_rate: float):
        assert new_flow_rate > 0
        self._blow_out_flow_rate = new_flow_rate

    @property
    def working_volume(self) -> float:
        """ The working volume of the pipette """
        return self._working_volume

    @working_volume.setter
    def working_volume(self, tip_volume: float):
        """ The working volume is the current tip max volume """
        self._working_volume = min(self.config.max_volume, tip_volume)

    @property
    def available_volume(self) -> float:
        """ The amount of liquid possible to aspirate """
        return self.working_volume - self.current_volume

    def set_current_volume(self, new_volume: float):
        assert new_volume >= 0
        assert new_volume <= self.working_volume
        self._current_volume = new_volume

    def add_current_volume(self, volume_incr: float):
        assert self.ok_to_add_volume(volume_incr)
        self._current_volume += volume_incr

    def remove_current_volume(self, volume_incr: float):
        assert self._current_volume >= volume_incr
        self._current_volume -= volume_incr

    def ok_to_add_volume(self, volume_incr: float) -> bool:
        return self.current_volume + volume_incr <= self.working_volume

    def add_tip(self,
                tip_length: float) -> None:
        """
        Add a tip to the pipette for position tracking and validation
        (effectively updates the pipette's critical point)

        :param tip_length: a positive, non-zero float presenting the distance
            in Z from the end of the pipette nozzle to the end of the tip
        :return:
        """
        assert tip_length > 0.0, "tip_length must be greater than 0"
        assert not self.has_tip
        self._has_tip = True
        self._current_tip_length = tip_length

    def remove_tip(self) -> None:
        """
        Remove the tip from the pipette (effectively updates the pipette's
        critical point)
        """
        assert self.has_tip
        self._has_tip = False
        self._current_tip_length = 0.0

    @property
    def has_tip(self) -> bool:
        return self._has_tip

    def ul_per_mm(self, ul: float, action: UlPerMmAction) -> float:
        sequence = self._config.ul_per_mm[action]
        return pipette_config.piecewise_volume_conversion(ul, sequence)

    def __str__(self) -> str:
        return '{} current volume {}ul critical point: {} at {}'\
            .format(self._config.display_name,
                    self.current_volume,
                    'tip end' if self.has_tip else 'nozzle end',
                    0)

    def __repr__(self) -> str:
        return '<{}: {} {}>'.format(self.__class__.__name__,
                                    self._config.display_name,
                                    id(self))

    def as_dict(self) -> 'Pipette.DictType':
        self._config_as_dict.update({
            'current_volume': self.current_volume,
            'available_volume': self.available_volume,
            'name': self.name,
            'model': self.model,
            'pipette_id': self.pipette_id,
            'has_tip': self.has_tip,
            'working_volume': self.working_volume,
            'aspirate_flow_rate':  self.aspirate_flow_rate,
            'dispense_flow_rate': self.dispense_flow_rate,
            'blow_out_flow_rate': self.blow_out_flow_rate
        })
        return self._config_as_dict


def _reload_and_check_skip(
        new_config: pipette_config.PipetteConfig,
        attached_instr: Pipette,
        pipette_offset: PipetteOffsetByPipetteMount) -> Tuple[Pipette, bool]:
    # Once we have determined that the new and attached pipettes
    # are similar enough that we might skip, see if the configs
    # match closely enough.
    # Returns a pipette object and True if we may skip hw reconfig
    if new_config == attached_instr.config\
       and pipette_offset == attached_instr._pipette_offset:
        # Same config, good enough
        return attached_instr, True
    else:
        newdict = asdict(new_config)
        olddict = asdict(attached_instr.config)
        changed: Set[str] = set()
        for k in newdict.keys():
            if newdict[k] != olddict[k]:
                changed.add(k)
        if changed.intersection(RECONFIG_KEYS):
            # Something has changed that requires reconfig
            p = Pipette(new_config,
                        pipette_offset,
                        attached_instr._pipette_id)
            p.act_as(attached_instr.acting_as)
            return p, False
    # Good to skip
    return attached_instr, True


def load_from_config_and_check_skip(
        config: Optional[pipette_config.PipetteConfig],
        attached: Optional[Pipette],
        requested: Optional[PipetteName],
        serial: Optional[str],
        pipette_offset: PipetteOffsetByPipetteMount)\
        -> Tuple[Optional[Pipette], bool]:
    """
    Given the pipette config for an attached pipette (if any) freshly read
    from disk, and any attached instruments,

    - Compare the new and configured pipette configs
    - Load the new configs if they differ
    - Return a bool indicating whether hardware reconfiguration may be
      skipped
    """

    if not config and not attached:
        # nothing attached now, nothing used to be attached, nothing
        # to reconfigure
        return attached, True

    if config and attached:
        # something was attached and something is attached. are they
        # the same? we can tell by comparing serials
        if serial == attached.pipette_id:
            if requested:
                # if there is an explicit instrument request, in addition
                # to checking if the old and new responses are the same
                # we also have to make sure the old pipette is properly
                # configured to the request
                if requested == attached.acting_as:
                    # similar enough to check
                    return _reload_and_check_skip(
                        config, attached, pipette_offset)
            else:
                # if there is no request, make sure that the old pipette
                # did not have backcompat applied
                if attached.acting_as == attached.name:
                    # similar enough to check
                    return _reload_and_check_skip(
                        config, attached, pipette_offset)

    if config:
        return \
            Pipette(config, pipette_offset, serial), False
    else:
        return None, False


def _build_splits(pipette: Pipette) -> Optional[MoveSplit]:
    if 'needsUnstick' in pipette.config.quirks:
        return MoveSplit(
            split_distance=1,
            split_current=1.75,
            split_speed=1,
            after_time=1800,
            fullstep=True)
    else:
        return None


def generate_hardware_configs(
        pipette: Optional[Pipette],
        robot_config: RobotConfig,
        revision: BoardRevision) -> InstrumentHardwareConfigs:
    """
    Fuse robot and pipette configuration to generate commands to send to
    the motor driver if required
    """
    if pipette:
        return {
            'steps_per_mm': pipette.config.steps_per_mm,
            'home_pos': pipette.config.home_position,
            'max_travel': pipette.config.max_travel,
            'idle_current': pipette.config.idle_current,
            'splits': _build_splits(pipette)
        }
    else:
        dpcs = robot_config.default_pipette_configs
        return {
            'steps_per_mm': dpcs['stepsPerMM'],
            'home_pos': dpcs['homePosition'],
            'max_travel': dpcs['maxTravel'],
            'idle_current': robot_configs.current_for_revision(
                robot_config.low_current, revision)['B'],
            'splits': None
        }



"""Protocol file runner module.

This module is responsible for logic that interprets different
types of protocol files in order to execute their logic. Primary
responsibilities of this module are:

- Extract metadata from protocol files
- Translate protocol file commands into ProtocolEngine commands
- Dispatch ProtocolEngine commands to an engine instance
"""

from .create_file_runner import create_file_runner
from .abstract_file_runner import AbstractFileRunner
from .json_file_runner import JsonFileRunner
from .python_file_runner import PythonFileRunner
from .protocol_file import ProtocolFileType, ProtocolFile

__all__ = [
    # runner factory
    "create_file_runner",
    # runner interfaces
    "AbstractFileRunner",
    "JsonFileRunner",
    "PythonFileRunner",
    # value objects
    "ProtocolFileType",
    "ProtocolFile",
]



"""Protocol engine errors module."""


class ProtocolEngineError(RuntimeError):
    """Base Protocol Engine error class."""

    pass


class UnexpectedProtocolError(ProtocolEngineError):
    """An error raised when an unexpected error occurs.

    This error is indicative of a software bug. If it happens, it means an
    exception was raised somewhere in the stack and it was not properly caught
    and wrapped.
    """

    def __init__(self, original_error: Exception) -> None:
        """Initialize an UnexpectedProtocolError with an original error."""
        super().__init__(str(original_error))
        self.original_error: Exception = original_error


class FailedToLoadPipetteError(ProtocolEngineError):
    """An error raised when executing a LoadPipetteRequest fails.

    This failure may be caused by:
    - An incorrect pipette already attached to the mount
    - A missing pipette on the requested mount
    """

    # TODO(mc, 2020-10-18): differentiate between pipette missing vs incorrect
    pass


class PipetteNotAttachedError(ProtocolEngineError):
    """An error raised when an operation's required pipette is not attached."""

    # TODO(mc, 2020-10-18): differentiate between pipette missing vs incorrect
    pass


class LabwareDoesNotExistError(ProtocolEngineError):
    """An error raised when referencing a labware that does not exist."""

    pass


class LabwareDefinitionDoesNotExistError(ProtocolEngineError):
    """An error raised when referencing a labware definition that does not exist."""

    pass


class LabwareIsNotTipRackError(ProtocolEngineError):
    """An error raised when trying to use a regular labware as a tip rack."""

    pass


class WellDoesNotExistError(ProtocolEngineError):
    """An error raised when referencing a well that does not exist."""

    pass


class PipetteDoesNotExistError(ProtocolEngineError):
    """An error raised when referencing a pipette that does not exist."""

    pass


class SlotDoesNotExistError(ProtocolEngineError):
    """An error raised when referencing a deck slot that does not exist."""

    pass


# TODO(mc, 2020-11-06): flesh out with structured data to replicate
# existing LabwareHeightError
class FailedToPlanMoveError(ProtocolEngineError):
    """An error raised when a requested movement could not be planned."""

    pass



import logging
from typing import Optional, Any

from opentrons import types
from opentrons.calibration_storage import get
from opentrons.calibration_storage.types import TipLengthCalNotFound
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocol_api.labware import Labware, Well
from opentrons.protocols.api_support.types import APIVersion
from opentrons_shared_data.protocol.dev_types import LiquidHandlingCommand, \
    BlowoutLocation


def validate_blowout_location(
        api_version: APIVersion,
        liquid_handling_command: LiquidHandlingCommand,
        blowout_location: Optional[Any]) -> None:
    """Validate the blowout location."""
    if blowout_location and api_version < APIVersion(2, 8):
        raise ValueError(
            'Cannot specify blowout location when using api' +
            ' version below 2.8, current version is {api_version}'
            .format(api_version=api_version))

    elif liquid_handling_command == 'consolidate' \
            and blowout_location == 'source well':
        raise ValueError(
            "blowout location for consolidate cannot be source well")
    elif liquid_handling_command == 'distribute' \
            and blowout_location == 'destination well':
        raise ValueError(
            "blowout location for distribute cannot be destination well")
    elif liquid_handling_command == 'transfer' and \
            blowout_location and \
            blowout_location not in \
            [location.value for location in BlowoutLocation]:
        raise ValueError(
            "blowout location should be either 'source well', " +
            " 'destination well', or 'trash'" +
            f" but it is {blowout_location}")


def tip_length_for(pipette: PipetteDict, tiprack: Labware) -> float:
    """ Get the tip length, including overlap, for a tip from this rack """

    def _build_length_from_overlap() -> float:
        tip_overlap = pipette['tip_overlap'].get(
            tiprack.uri,
            pipette['tip_overlap']['default'])
        tip_length = tiprack.tip_length
        return tip_length - tip_overlap

    try:
        return get.load_tip_length_calibration(
            pipette['pipette_id'],
            tiprack._implementation.get_definition()
            ).tip_length
    except TipLengthCalNotFound:
        return _build_length_from_overlap()


VALID_PIP_TIPRACK_VOL = {
    'p10': [10, 20],
    'p20': [10, 20],
    'p50': [200, 300],
    'p300': [200, 300],
    'p1000': [1000]
}


def validate_tiprack(
        instrument_name: str,
        tiprack: Labware,
        log: logging.Logger) -> None:
    """Validate a tiprack logging a warning message."""
    # TODO AA 2020-06-24 - we should instead add the acceptable Opentrons
    #  tipracks to the pipette as a refactor
    if tiprack._implementation.get_definition()['namespace'] \
            == 'opentrons':
        tiprack_vol = tiprack.wells()[0].max_volume
        valid_vols = VALID_PIP_TIPRACK_VOL[instrument_name.split('_')[0]]
        if tiprack_vol not in valid_vols:
            log.warning(
                f'The pipette {instrument_name} and its tiprack '
                f'{tiprack.load_name} in slot {tiprack.parent} appear to '
                'be mismatched. Please check your protocol before running '
                'on the robot.')


def determine_drop_target(
        api_version: APIVersion,
        location: Well,
        return_height: float,
        version_breakpoint: APIVersion = None) -> types.Location:
    """Determine the drop target based on well and api version."""
    version_breakpoint = version_breakpoint or APIVersion(2, 2)
    if api_version < version_breakpoint:
        bot = location.bottom()
        return types.Location(
            point=bot.point._replace(z=bot.point.z + 10),
            labware=location)
    else:
        tr = location.parent
        assert tr.is_tiprack
        z_height = return_height * tr.tip_length
        return location.top(-z_height)


def validate_can_aspirate(location: types.Location) -> None:
    """ Can one aspirate on the given `location` or not? This method is
    pretty basic and will probably remain so (?) as the future holds neat
    ambitions for how validation is implemented. And as robots become more
    intelligent more rigorous testing will be possible

    Args:
        location: target for aspiration

    Raises:
        RuntimeError:
    """
    if _is_tiprack(location):
        raise RuntimeError("Cannot aspirate a tiprack")


def validate_can_dispense(location: types.Location) -> None:
    """ Can one dispense to the given `location` or not? This method is
    pretty basic and will probably remain so (?) as the future holds neat
    ambitions for how validation is implemented. And as robots become more
    intelligent more rigorous testing will be possible

    Args:
        location: target for dispense

    Raises:
        RuntimeError:
    """
    if _is_tiprack(location):
        raise RuntimeError("Cannot dispense to a tiprack")


def _is_tiprack(location: types.Location) -> bool:
    labware = location.labware.as_labware()
    return labware.parent and labware.parent.is_tiprack



from .labware_definition import (
    LabwareDefinition,
    WellDefinition
)
from .json_protocol import Model as JsonProtocol

__all__ = [
    "LabwareDefinition",
    "WellDefinition",
    "JsonProtocol",
]



from typing_extensions import Final


HOMED_POSITION: Final = {
    'X': 418.0,
    'Y': 353.0,
    'Z': 218.0,
    'A': 218.0,
    'B': 19.0,
    'C': 19.0
}
Y_BOUND_OVERRIDE: Final = 370


class SmoothieDriver(object):

    def __init__(self):
        pass


class VirtualSmoothie(object):

    def __init__(self):
        pass


class SimulatingDriver:
    def __init__(self):
        self._steps_per_mm = {}

    def home(self, axis):
        pass

    def _smoothie_reset(self):
        pass

    def read_pipette_id(self, mount):
        pass

    def read_pipette_model(self, mount):
        pass

    def write_pipette_id(self, mount, id):
        pass

    def write_pipette_model(self, mount, model):
        pass

    def _send_command(self, command, timeout=None):
        pass

    def turn_on_blue_button_light(self):
        pass

    def turn_on_red_button_light(self):
        pass

    def update_pipette_config(self, axis, data):
        '''
        Updates the following configs for a given pipette mount based on
        the detected pipette type:
        - homing positions M365.0
        - Max Travel M365.1
        - endstop debounce M365.2 (NOT for zprobe debounce)
        - retract from endstop distance M365.3
        '''
        pass

    @property
    def current(self):
        pass

    @property
    def speed(self):
        pass

    @property
    def steps_per_mm(self):
        return self._steps_per_mm

    @steps_per_mm.setter
    def steps_per_mm(self, axis, mm):
        # Keep track of any updates to the steps per mm per axis
        self._steps_per_mm[axis] = mm

    def update_steps_per_mm(self, data):
        pass

    def configure_splits_for(self, config):
        pass

    def set_dwelling_current(self, settings):
        pass

    def set_acceleration(self, settings):
        pass

    @property
    def homed_position(self):
        return HOMED_POSITION.copy()

    @property
    def axis_bounds(self):
        position = HOMED_POSITION.copy()
        position['Y'] = Y_BOUND_OVERRIDE
        return position



import typing

from opentrons import types
from opentrons.hardware_control import NoTipAttachedError, TipAttachedError
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.types import HardwareAction
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.api_support.util import FlowRates, PlungerSpeeds, \
    Clearances
from opentrons.protocols.geometry import planning
from opentrons.protocols.context.instrument import \
    AbstractInstrument
from opentrons.protocols.context.protocol import \
    AbstractProtocol
from opentrons.protocols.context.well import WellImplementation


class InstrumentContextSimulation(AbstractInstrument):
    """A simulation of an instrument context."""

    def __init__(self,
                 protocol_interface: AbstractProtocol,
                 pipette_dict: PipetteDict,
                 mount: types.Mount,
                 instrument_name: str,
                 default_speed: float = 400.0):
        """Constructor."""
        self._protocol_interface = protocol_interface
        self._mount = mount
        self._pipette_dict = pipette_dict
        self._instrument_name = instrument_name
        self._default_speed = default_speed
        self._flow_rate = FlowRates(self)
        self._plunger_speeds = PlungerSpeeds(self)
        # Cache the maximum instrument height
        self._instrument_max_height = protocol_interface\
            .get_hardware().hardware.get_instrument_max_height(self._mount)

    def get_default_speed(self) -> float:
        return self._default_speed

    def set_default_speed(self, speed: float) -> None:
        self._default_speed = speed

    def aspirate(self, volume: float, rate: float) -> None:
        self._raise_if_no_tip(HardwareAction.ASPIRATE.name)
        new_volume = self.get_current_volume() + volume
        assert new_volume <= self._pipette_dict['working_volume'],\
            "Cannot aspirate more than pipette max volume"
        self._pipette_dict['current_volume'] = new_volume

    def dispense(self, volume: float, rate: float) -> None:
        self._raise_if_no_tip(HardwareAction.DISPENSE.name)
        self._pipette_dict['current_volume'] -= volume

    def blow_out(self) -> None:
        self._raise_if_no_tip(HardwareAction.BLOWOUT.name)
        self._pipette_dict['current_volume'] = 0
        self._pipette_dict['ready_to_aspirate'] = False

    def touch_tip(self, location: WellImplementation, radius: float,
                  v_offset: float, speed: float) -> None:
        pass

    def pick_up_tip(self, well: WellImplementation, tip_length: float,
                    presses: typing.Optional[int],
                    increment: typing.Optional[float]) -> None:
        self._raise_if_tip("drop tip")
        self._pipette_dict['has_tip'] = True
        self._pipette_dict['current_volume'] = 0

    def drop_tip(self, home_after: bool) -> None:
        self._raise_if_no_tip(HardwareAction.DROPTIP.name)
        self._pipette_dict['has_tip'] = False

    def home(self) -> None:
        self._protocol_interface.set_last_location(None)

    def home_plunger(self) -> None:
        pass

    def delay(self) -> None:
        pass

    def move_to(self, location: types.Location, force_direct: bool,
                minimum_z_height: typing.Optional[float],
                speed: typing.Optional[float]) -> None:
        """Simulation of only the motion planning portion of move_to."""
        last_location = self._protocol_interface.get_last_location()
        if last_location:
            from_loc = last_location
        else:
            from_loc = types.Location(types.Point(0, 0, 0), LabwareLike(None))

        # We just want to catch planning errors.
        planning.plan_moves(from_loc,
                            location,
                            self._protocol_interface.get_deck(),
                            self._instrument_max_height,
                            force_direct=force_direct,
                            minimum_z_height=minimum_z_height)

        self._protocol_interface.set_last_location(location)

    def get_mount(self) -> types.Mount:
        return self._mount

    def get_instrument_name(self) -> str:
        return self._instrument_name

    def get_pipette_name(self) -> str:
        return self._pipette_dict['name']

    def get_model(self) -> str:
        return self._pipette_dict['model']

    def get_min_volume(self) -> float:
        return self._pipette_dict['min_volume']

    def get_max_volume(self) -> float:
        return self._pipette_dict['max_volume']

    def get_current_volume(self) -> float:
        return self._pipette_dict['current_volume']

    def get_available_volume(self) -> float:
        return self._pipette_dict['available_volume']

    def get_pipette(self) -> PipetteDict:
        return self._pipette_dict

    def get_channels(self) -> int:
        return self._pipette_dict['channels']

    def has_tip(self) -> bool:
        return self._pipette_dict['has_tip']

    def is_ready_to_aspirate(self) -> bool:
        return self._pipette_dict['ready_to_aspirate']

    def prepare_for_aspirate(self) -> None:
        self._raise_if_no_tip(HardwareAction.PREPARE_ASPIRATE.name)

    def get_return_height(self) -> float:
        return self._pipette_dict['return_tip_height']

    def get_well_bottom_clearance(self) -> Clearances:
        return Clearances(default_aspirate=1, default_dispense=1)

    def get_speed(self) -> PlungerSpeeds:
        return self._plunger_speeds

    def get_flow_rate(self) -> FlowRates:
        return self._flow_rate

    def set_flow_rate(self, aspirate: typing.Optional[float] = None,
                      dispense: typing.Optional[float] = None,
                      blow_out: typing.Optional[float] = None) -> None:
        if aspirate is not None:
            self._pipette_dict['aspirate_flow_rate'] = aspirate
        if dispense is not None:
            self._pipette_dict['dispense_flow_rate'] = dispense
        if blow_out is not None:
            self._pipette_dict['blow_out_flow_rate'] = blow_out

    def set_pipette_speed(self, aspirate: typing.Optional[float] = None,
                          dispense: typing.Optional[float] = None,
                          blow_out: typing.Optional[float] = None) -> None:
        if aspirate is not None:
            self._pipette_dict['aspirate_speed'] = aspirate
        if dispense is not None:
            self._pipette_dict['dispense_speed'] = dispense
        if blow_out is not None:
            self._pipette_dict['blow_out_speed'] = blow_out

    def _raise_if_no_tip(self, action: str) -> None:
        """Raise NoTipAttachedError if no tip."""
        if not self.has_tip():
            raise NoTipAttachedError(
                f'Cannot perform {action} without a tip attached'
            )

    def _raise_if_tip(self, action: str) -> None:
        """Raise TipAttachedError if tip."""
        if self.has_tip():
            raise TipAttachedError(
                f'Cannot {action} with a tip attached'
            )



"""Value objects and models representing protocol files."""
# TODO(mc, 2021-04-30): as these objects are fleshed out, pull in
# existing logic and models from:
#   - api/src/opentrons/protocols/types.py
#   - robot-server/robot_server/service/protocol/models.py
from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from pathlib import Path


class ProtocolFileType(str, Enum):
    """Type of a protocol file.

    Attributes:
        PYTHON: a Python protocol file or module
        JSON: a JSON protocol
    """

    PYTHON = "python"
    JSON = "json"


@dataclass(frozen=True)
class ProtocolFile:
    """A value object representing a protocol file on disk.

    Attributes:
        file_type: Whether the file is a JSON protocol or Python protocol
    """

    file_path: Path
    file_type: ProtocolFileType



from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Sequence
from opentrons.protocols.context.well import WellImplementation


Wells = Sequence[WellImplementation]
HeaderToWells = Dict[str, Wells]
WellsByDimension = Sequence[Wells]


@dataclass
class Grid:
    rows: HeaderToWells
    columns: HeaderToWells


class WellGrid:
    """A helper class to extract Wells by row or column"""

    def __init__(self, wells: Wells):
        """
        Construct well grid from a collection of well objects ordered as they
         appear in ordering` field of Labware Defnition
        """
        self._grid = self._create_row_column(wells)
        self._row_headers = sorted(self._grid.rows.keys())
        self._column_headers = sorted(self._grid.columns.keys(),
                                      key=lambda k: int(k))
        self._rows = [self._grid.rows[h] for h in self._row_headers]
        self._columns = [self._grid.columns[h] for h in self._column_headers]

    def row_headers(self) -> Sequence[str]:
        """List of row header names"""
        return self._row_headers

    def column_headers(self) -> Sequence[str]:
        """List of column header names"""
        return self._column_headers

    def get_row_dict(self) -> HeaderToWells:
        """A mapping of row header to a list"""
        return self._grid.rows

    def get_column_dict(self) -> HeaderToWells:
        """A mapping of column header to a list"""
        return self._grid.columns

    def get_rows(self) -> WellsByDimension:
        """Get all rows as list of lists"""
        return self._rows

    def get_columns(self) -> WellsByDimension:
        """Get all columns as list of lists"""
        return self._columns

    def get_row(self, row: str) -> Wells:
        """Get an individual row"""
        return self._grid.rows.get(row, [])

    def get_column(self, column: str) -> Wells:
        """Get an individual column"""
        return self._grid.columns.get(column, [])

    @staticmethod
    def _create_row_column(wells: Wells) -> Grid:
        """
        Creates a dict of lists of Wells. Which way the labware is segmented
        determines whether this is a dict of rows or dict of columns. If group
        is 1, then it will collect wells that have the same alphabetic prefix
        and therefore are considered to be in the same row. If group is 2, it
        will collect wells that have the same numeric postfix and therefore
        are considered to be in the same column.
        """
        columns = defaultdict(list)
        rows = defaultdict(list)
        for well in wells:
            rows[well.get_row_name()].append(well)
            columns[well.get_column_name()].append(well)
        # copy to a non-default-dict
        return Grid(columns=dict(columns), rows=dict(rows))



from .protocol_context import ProtocolContext
from .instrument_context import InstrumentContext
from .paired_instrument_context import PairedInstrumentContext
from .module_contexts import (
    ModuleContext, ThermocyclerContext, MagneticModuleContext,
    TemperatureModuleContext)


__all__ = [
    'ProtocolContext', 'InstrumentContext', 'ModuleContext', 'PairedInstrumentContext',
    'ThermocyclerContext', 'MagneticModuleContext', 'TemperatureModuleContext'
]



"""An emulation of the Smoothie.

The purpose is to provide a fake backend that responds to the GCODE sent by the
Opentrons smoothie driver.
"""
import logging
import re
from typing import Optional, Dict

from opentrons import _find_smoothie_file
from opentrons.drivers import utils
from opentrons.drivers.smoothie_drivers import HOMED_POSITION
from opentrons.drivers.smoothie_drivers.driver_3_0 import GCODE
from opentrons.hardware_control.emulation.parser import Command, Parser

from .abstract_emulator import AbstractEmulator
from .settings import SmoothieSettings

logger = logging.getLogger(__name__)


class SmoothieEmulator(AbstractEmulator):
    """Smoothie emulator"""

    WRITE_INSTRUMENT_RE = re.compile(r"(?P<mount>[LR])\s*(?P<value>[a-f0-9]+)")

    def __init__(self, parser: Parser, settings: SmoothieSettings) -> None:
        """Constructor"""
        _, fw_version = _find_smoothie_file()
        self._version_string = \
            f"Build version: {fw_version}, Build date: CURRENT, " \
            f"MCU: NONE, System Clock: NONE"

        self._pos = {'A': 0.0, 'B': 0.0, 'C': 0.0, 'X': 0.0, 'Y': 0.0, 'Z': 0.0}
        self._home_status: Dict[str, bool] = {
            'X': False,
            'Y': False,
            'Z': False,
            'A': False,
            'B': False,
            'C': False,
        }
        self._speed = 0.0
        self._pipette_model = {
            "L": utils.string_to_hex(settings.left.model, 64),
            "R": utils.string_to_hex(settings.right.model, 64)
        }
        self._pipette_id = {
            "L": utils.string_to_hex(settings.left.id, 64),
            "R": utils.string_to_hex(settings.right.id, 64),
        }
        self._parser = parser

    def handle(self, line: str) -> Optional[str]:
        """Handle a line"""
        results = (self._handle(c) for c in self._parser.parse(line))
        joined = ' '.join(r for r in results if r)
        return None if not joined else joined

    def _handle(self, command: Command) -> Optional[str]:  # noqa: C901
        """Handle a command."""
        # TODO (al, 2021-04-28): break this up into multiple functions and
        #  remove 'noqa(C901)'.
        logger.info(f"Got command {command}")
        if GCODE.HOMING_STATUS == command.gcode:
            vals = " ".join(f"{k}:{int(v)}" for k, v in self._home_status.items())
            return vals
        elif GCODE.CURRENT_POSITION == command.gcode:
            vals = " ".join(f"{k}:{v}" for k, v in self._pos.items())
            return f"{command.gcode}\r\n\r\nok MCS: {vals}"
        elif GCODE.VERSION == command.gcode:
            return self._version_string
        elif GCODE.READ_INSTRUMENT_ID == command.gcode:
            if "L" in command.params:
                return f"L:{self._pipette_id['L']}"
            elif "R" in command.params:
                return f"R:{self._pipette_id['R']}"
        elif GCODE.READ_INSTRUMENT_MODEL == command.gcode:
            if "L" in command.params:
                return f"L:{self._pipette_model['L']}"
            elif "R" in command.params:
                return f"R:{self._pipette_model['R']}"
        elif GCODE.WRITE_INSTRUMENT_ID == command.gcode:
            self._pipette_id.update(self._mount_strings(command))
        elif GCODE.WRITE_INSTRUMENT_MODEL == command.gcode:
            self._pipette_model.update(self._mount_strings(command))
        elif GCODE.MOVE == command.gcode:
            for key, value in command.params.items():
                assert isinstance(value, float), f"invalid value '{value}'"
                if 'F' == key:
                    self._speed = value
                else:
                    self._pos[key] = value
        elif GCODE.HOME == command.gcode:
            for axis in command.params.keys():
                self._pos[axis] = HOMED_POSITION[axis]
                self._home_status[axis] = True
        return None

    @staticmethod
    def _mount_strings(command: Command) -> Dict[str, str]:
        """
        Parse the body of the command for the mount strings.

        Write instrument id and model use the format:
            MOUNT VALUE
        where MOUNT is either L or R and VALUE is a string of ascii codes.

        Args:
            command: Command

        Returns:
            A dict of L and/or R to the string value following it.
        """
        pars = (i.groupdict() for i in
                SmoothieEmulator.WRITE_INSTRUMENT_RE.finditer(command.body))
        result = {p['mount']: p['value'] for p in pars}
        assert result, f"missing mount values '{command.body}'"
        return result



from __future__ import annotations

from typing import cast, Tuple, TYPE_CHECKING, List, Union, Optional
from opentrons.types import Location
from opentrons.protocols.api_support import labware_like

import opentrons.commands.types as command_types

if TYPE_CHECKING:
    from opentrons.protocol_api import InstrumentContext
    from opentrons.protocol_api.labware import Labware, Well
    from opentrons.protocols.geometry.module_geometry import ModuleGeometry
    ReferredObjects = Tuple[
        List[InstrumentContext], List[Labware], List[ModuleGeometry]]


def _labware_from_location(loc: Union[Location, Well]) -> Labware:
    if isinstance(loc, Location):
        if not loc.labware:
            raise ValueError('Cannot handle location without labware')
        labware, _ = loc.labware.get_parent_labware_and_well()
        if not labware:
            raise ValueError(f'Cannot handle location {loc}')
        return labware
    else:
        return loc.parent

# The casts and type: ignores in _get_locations and _get_instruments are necessary
# until we either a) refactor this so payloads have tags that mypy can use to
# discriminate the unions (or make commands dataclasses that can be checked with
# isinstance) or b) mypy learns full disjoint union discrimination.


def _get_locations(
        command: command_types.HasLocationPayload) -> List[Optional[Labware]]:
    if 'location' in command and command['location']:  # type: ignore
        return [_labware_from_location(
            cast(command_types.SingleLocationPayload, command)['location'])]
    elif 'locations' in command and command['locations']:  # type: ignore
        return [_labware_from_location(loc) for loc in
                cast(command_types.MultiLocationPayload, command)['locations']]
    else:
        return []


def _get_instruments(
        command: command_types.HasInstrumentPayload) -> List[InstrumentContext]:
    if 'instrument' in command:  # type: ignore
        return [cast(command_types.SingleInstrumentPayload, command)['instrument']]
    else:  # type: ignore
        return [
            inst for inst
            in cast(command_types.MultiInstrumentPayload, command)['instruments']]


def get_referred_objects(command: command_types.CommandPayload) -> ReferredObjects:
    if 'location' in command or 'locations' in command:
        locations = [lw for lw in _get_locations(
            cast(command_types.HasLocationPayload, command)) if lw is not None]
    else:
        locations = []
    if 'instrument' in command or 'instruments' in command:
        instruments = _get_instruments(
            cast(command_types.HasInstrumentPayload, command))
    else:
        instruments = []

    maybe_modules = [
        labware_like.LabwareLike(labware).module_parent() for labware in locations]
    return instruments, locations, [mod for mod in maybe_modules if mod]



"""Base state store classes."""
from abc import ABC
from typing import Generic, TypeVar

from ..commands import CompletedCommandType


SubstateT = TypeVar("SubstateT")


class Substore(ABC, Generic[SubstateT]):
    """Abstract base class for a sub-store."""

    _state: SubstateT

    @property
    def state(self) -> SubstateT:
        """State getter."""
        return self._state


class CommandReactive(ABC):
    """Abstract base class for an interface that reacts to commands."""

    def handle_completed_command(self, command: CompletedCommandType) -> None:
        """React to a CompletedCommand."""
        pass



"""Labware data resource provider."""
from typing import Tuple, cast
from opentrons_shared_data.labware import dev_types
from opentrons.protocols.models import LabwareDefinition
from opentrons.protocol_api.labware import get_labware_definition
from opentrons.calibration_storage.get import get_labware_calibration
from opentrons.calibration_storage.helpers import hash_labware_def

from ..types import LabwareLocation


class LabwareDataProvider:
    """Labware data provider."""

    # NOTE(mc, 2020-10-18): async to allow file reading and parsing to be
    # async on a worker thread in the future
    async def get_labware_definition(
        self,
        load_name: str,
        namespace: str,
        version: int,
    ) -> LabwareDefinition:
        """Get a labware definition given the labware's identification."""
        return LabwareDefinition.parse_obj(
            get_labware_definition(load_name, namespace, version)
        )

    # NOTE(mc, 2020-10-18): async to allow file reading and parsing to be
    # async on a worker thread in the future
    async def get_labware_calibration(
        self,
        definition: LabwareDefinition,
        location: LabwareLocation,
    ) -> Tuple[float, float, float]:
        """Get a labware's calibration data given its definition."""
        # TODO(mc, 2020-10-18): Fetching labware calibration data is a little
        #  convoluted and could use some clean up
        as_type_dict = cast(dev_types.LabwareDefinition, definition.dict())
        labware_path = f'{hash_labware_def(as_type_dict)}.json'
        cal_point = get_labware_calibration(
            labware_path,
            as_type_dict,
            # TODO(mc, 2020-10-18): Support labware on modules
            parent='',
        )
        return (cal_point.x, cal_point.y, cal_point.z)



from os import environ
import logging
import asyncio
from threading import Event, Thread, Lock
from time import sleep
from typing import Any, Optional, Mapping, Dict, Tuple
from serial.serialutil import SerialException  # type: ignore

from opentrons.drivers import serial_communication, utils
from opentrons.drivers.serial_communication import SerialNoResponse
from opentrons.drivers.types import Temperature

"""
- Driver is responsible for providing an interface for the temp-deck
- Driver is the only system component that knows about the temp-deck's GCODES
  or how the temp-deck communications

- Driver is NOT responsible interpreting the temperatures or states in any way
  or knowing anything about what the device is being used for
"""

log = logging.getLogger(__name__)

ERROR_KEYWORD = 'error'
ALARM_KEYWORD = 'alarm'

DEFAULT_TEMP_DECK_TIMEOUT = 1

DEFAULT_STABILIZE_DELAY = 0.1
DEFAULT_COMMAND_RETRIES = 3

GCODES = {
    'GET_TEMP': 'M105',
    'SET_TEMP': 'M104',
    'DEVICE_INFO': 'M115',
    'DISENGAGE': 'M18',
    'PROGRAMMING_MODE': 'dfu'
}

TEMP_DECK_BAUDRATE = 115200

TEMP_DECK_COMMAND_TERMINATOR = '\r\n\r\n'
TEMP_DECK_ACK = 'ok\r\nok\r\n'

TEMP_DECK_MODELS = {
    'temperatureModuleV1': 'temp_deck_v1.1',
    'temperatureModuleV2': 'temp_deck_v20'
}

temp_locks: Dict[str, Tuple[Lock, 'TempDeck']] = {}


class TempDeckError(Exception):
    pass


class SimulatingDriver:
    def __init__(self, sim_model: str = None):
        self._target_temp = 0.0
        self._active = False
        self._port: Optional[str] = None
        self._model = TEMP_DECK_MODELS[sim_model] if sim_model\
            else 'temp_deck_v1.1'

    async def set_temperature(self, celsius: float):
        self._target_temp = celsius
        self._active = True

    def start_set_temperature(self, celsius):
        self._target_temp = celsius
        self._active = True

    def legacy_set_temperature(self, celsius: float):
        self._target_temp = celsius
        self._active = True

    def deactivate(self):
        self._target_temp = 0
        self._active = False

    def update_temperature(self):
        pass

    def connect(self, port: str):
        self._port = port

    def is_connected(self) -> bool:
        return True

    def disconnect(self):
        pass

    def enter_programming_mode(self):
        pass

    @property
    def temperature(self) -> float:
        return self._target_temp

    @property
    def target(self) -> Optional[float]:
        return self._target_temp if self._active else None

    @property
    def status(self) -> str:
        return 'holding at target' if self._active else 'idle'

    def get_device_info(self) -> Mapping[str, str]:
        return {'serial': 'dummySerialTD',
                'model': self._model,
                'version': 'dummyVersionTD'}


class TempDeck:
    def __init__(self, config={}):
        self.run_flag = Event()
        self.run_flag.set()

        self._connection = None
        self._config = config

        self._temperature = Temperature(current=25, target=None)
        self._update_thread = None
        self._port = None
        self._lock = None

    def connect(self, port=None) -> Optional[str]:
        if environ.get('ENABLE_VIRTUAL_SMOOTHIE', '').lower() == 'true':
            return None
        try:
            self.disconnect(port)
            self._connect_to_port(port)
            if temp_locks.get(port):
                self._lock = temp_locks[port][0]
            else:
                self._lock = Lock()
                temp_locks[port] = (self._lock, self)
            self._wait_for_ack()  # verify the device is there
            self._port = port

        except (SerialException, SerialNoResponse) as e:
            return str(e)
        return ''

    def disconnect(self, port=None):
        if self._port and self.is_connected():
            self._connection.close()  # type: ignore
            del temp_locks[self._port]
        elif self.is_connected():
            self._connection.close()  # type: ignore

        self._connection = None

    def is_connected(self) -> bool:
        if not self._connection:
            return False
        return self._connection.is_open

    @property
    def port(self) -> Optional[str]:
        if not self._connection:
            return None
        return self._connection.port

    def deactivate(self) -> str:
        self.run_flag.wait()
        try:
            self._send_command(GCODES['DISENGAGE'])
        except (TempDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        return ''

    async def set_temperature(self, celsius) -> str:
        self.run_flag.wait()
        celsius = round(float(celsius),
                        utils.TEMPDECK_GCODE_ROUNDING_PRECISION)
        try:
            self._send_command(
                '{0} S{1}'.format(GCODES['SET_TEMP'], celsius))
        except (TempDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        self._temperature.target = celsius
        while self.status != 'holding at target':
            await asyncio.sleep(0.1)
        return ''

    def start_set_temperature(self, celsius) -> str:
        self.run_flag.wait()
        celsius = round(float(celsius),
                        utils.TEMPDECK_GCODE_ROUNDING_PRECISION)
        self._send_command(
                '{0} S{1}'.format(GCODES['SET_TEMP'], celsius))
        self._temperature.target = celsius
        return ''

    def update_temperature(self) -> str:
        if self._update_thread and self._update_thread.is_alive():
            # No need to do anything. The updater is already running.
            pass
        else:
            def _update():
                try:
                    self._recursive_update_temperature(retries=DEFAULT_COMMAND_RETRIES)
                except (OSError, TempDeckError, SerialException, SerialNoResponse):
                    log.exception("Failed to execute _recursive_update_temperature.")

            self._update_thread = Thread(
                target=_update(),
                name='Tempdeck recursive update temperature')
            self._update_thread.start()
        return ''

    @property
    def target(self) -> Optional[float]:
        return self._temperature.target

    @property
    def temperature(self) -> float:
        return self._temperature.current  # type: ignore

    def _get_status(self) -> str:
        # Separate function for testability
        current = self._temperature.current
        target = self._temperature.target
        delta = 0.7
        if target:
            diff = target - current  # type: ignore
            if abs(diff) < delta:   # To avoid status fluctuation near target
                return 'holding at target'
            elif diff < 0:
                return 'cooling'
            else:
                return 'heating'
        else:
            return 'idle'

    @property
    def status(self) -> str:
        return self._get_status()

    def get_device_info(self) -> Mapping[str, str]:
        """
        Queries Temp-Deck for its build version, model, and serial number

        returns: dict
            Where keys are the strings 'version', 'model', and 'serial',
            and each value is a string identifier

            {
                'serial': '1aa11bb22',
                'model': '1aa11bb22',
                'version': '1aa11bb22'
            }

        Example input from Temp-Deck's serial response:
            "serial:aa11bb22 model:aa11bb22 version:aa11bb22"
        """
        return self._get_info(DEFAULT_COMMAND_RETRIES)

    def pause(self):
        self.run_flag.clear()

    def resume(self):
        self.run_flag.set()

    def enter_programming_mode(self) -> str:
        try:
            self._send_command(GCODES['PROGRAMMING_MODE'])
        except (TempDeckError, SerialException, SerialNoResponse) as e:
            return str(e)
        if self._port:
            del temp_locks[self._port]
        return ''

    def _connect_to_port(self, port=None):
        try:
            self._connection = serial_communication.connect(
                port=port,
                baudrate=TEMP_DECK_BAUDRATE
            )
        except SerialException:
            # if another process is using the port, pyserial raises an
            # exception that describes a "readiness to read" which is confusing
            error_msg = 'Unable to access Serial port to Temp-Deck. This is '
            error_msg += 'because another process is currently using it, or '
            error_msg += 'the Serial port is disabled on this device (OS)'
            raise SerialException(error_msg)

    def _wait_for_ack(self):
        """
        This methods writes a sequence of newline characters, which will
        guarantee temp-deck responds with 'ok\r\nok\r\n' within 1 seconds
        """
        self._send_command('\r\n', timeout=DEFAULT_TEMP_DECK_TIMEOUT)

    # Potential place for command optimization (buffering, flushing, etc)
    def _send_command(
            self, command, timeout=DEFAULT_TEMP_DECK_TIMEOUT, tag=None):
        assert self._lock, 'not connected'
        with self._lock:
            command_line = command + ' ' + TEMP_DECK_COMMAND_TERMINATOR
            ret_code = self._recursive_write_and_return(
                command_line, timeout, DEFAULT_COMMAND_RETRIES)

            # Smoothieware returns error state if a switch was hit while moving
            if (ERROR_KEYWORD in ret_code.lower()) or \
                    (ALARM_KEYWORD in ret_code.lower()):
                log.error(f'Received error message from Temp-Deck: {ret_code}')
                raise TempDeckError(ret_code)

            return ret_code.strip()

    def _recursive_write_and_return(self, cmd, timeout, retries, tag=None):
        if not tag:
            tag = f'tempdeck {id(self)}'
        try:
            return serial_communication.write_and_return(
                cmd,
                TEMP_DECK_ACK,
                self._connection,
                timeout,
                tag=tag)
        except SerialNoResponse as e:
            retries -= 1
            if retries <= 0:
                raise e
            sleep(DEFAULT_STABILIZE_DELAY)
            if self._connection:
                self._connection.close()
                self._connection.open()
            return self._recursive_write_and_return(
                cmd, timeout, retries, tag=tag)

    def _recursive_update_temperature(self, retries):
        try:
            res = self._send_command(
                GCODES['GET_TEMP'],
                tag=f'tempdeck {id(self)} rut')
            self._temperature = utils.parse_temperature_response(
                res, utils.TEMPDECK_GCODE_ROUNDING_PRECISION)
            return None
        except utils.ParseError as e:
            retries -= 1
            if retries <= 0:
                raise TempDeckError(e)
            sleep(DEFAULT_STABILIZE_DELAY)
            return self._recursive_update_temperature(retries)

    def _get_info(self, retries) -> Mapping[str, str]:
        last_e: Any = None
        for _ in range(retries):
            try:
                device_info = self._send_command(GCODES['DEVICE_INFO'])
                return utils.parse_device_information(device_info)
            except utils.ParseError as e:
                log.exception("tempdeck device information parse failure")
                last_e = e
                sleep(DEFAULT_STABILIZE_DELAY)
        if last_e:
            raise last_e
        else:
            raise TempDeckError('Unknown error in temperature module')



""" opentrons.simulate: functions and entrypoints for simulating protocols

This module has functions that provide a console entrypoint for simulating
a protocol from the command line.
"""

import argparse
import asyncio

import sys
import logging
import os
import pathlib
import queue
from typing import (Any, Dict, List, Mapping, TextIO, Tuple, BinaryIO,
                    Optional, Union, TYPE_CHECKING)


import opentrons
from opentrons.hardware_control.simulator_setup import load_simulator
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocols.execution import execute
import opentrons.broker
from opentrons.config import IS_ROBOT, JUPYTER_NOTEBOOK_LABWARE_DIR
from opentrons import protocol_api
from opentrons.commands import types as command_types
from opentrons.protocols.api_support.util import HardwareToManage
from opentrons.protocols.context.protocol_api.protocol_context import \
    ProtocolContextImplementation
from opentrons.protocols import parse, bundle
from opentrons.protocols.types import (
    PythonProtocol, BundleContents)
from opentrons.protocols.api_support.types import APIVersion
from .util.entrypoint_util import labware_from_paths, datafiles_from_paths

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition


class AccumulatingHandler(logging.Handler):
    def __init__(self, level, command_queue):
        """ Create the handler

        :param level: The logging level to capture
        :param command_queue: The queue.Queue to use for messages
        """
        self._command_queue = command_queue
        super().__init__(level)

    def emit(self, record):
        self._command_queue.put(record)


class CommandScraper:
    """ An object that handles scraping the broker for commands

    This should be instantiated with the logger to integrate
    messages from (e.g. ``logging.getLogger('opentrons')``), the
    level to scrape, and the opentrons broker object to subscribe to.

    The :py:attr:`commands` property contains the list of commands
    and log messages integrated together. Each element of the list is
    a dict following the pattern in the docs of :py:meth:`simulate`.
    """

    def __init__(self,
                 logger: logging.Logger,
                 level: str,
                 broker: opentrons.broker.Broker) -> None:
        """ Build the scraper.

        :param logger: The :py:class:`logging.logger` to scrape
        :param level: The log level to scrape
        :param broker: Which broker to subscribe to
        """
        self._logger = logger
        self._broker = broker
        self._queue = queue.Queue()  # type: ignore
        if level != 'none':
            level = getattr(logging, level.upper(), logging.WARNING)
            self._logger.setLevel(level)
            self._handler: Optional[AccumulatingHandler]\
                = AccumulatingHandler(
                    level, self._queue)
            logger.addHandler(self._handler)
        else:
            self._handler = None
        self._depth = 0
        self._commands: List[Mapping[str, Any]] = []
        self._unsub = self._broker.subscribe(
            command_types.COMMAND,
            self._command_callback)

    @property
    def commands(self) -> List[Mapping[str, Mapping[str, Any]]]:
        """ The list of commands. See :py:meth:`simulate` """
        return self._commands

    def __del__(self):
        if getattr(self, '_handler', None):
            try:
                self._logger.removeHandler(self._handler)  # type: ignore
            except Exception:
                pass
        if hasattr(self, '_unsub'):
            self._unsub()

    def _command_callback(self, message):
        """ The callback subscribed to the broker """
        payload = message['payload']
        if message['$'] == 'before':
            self._commands.append({'level': self._depth,
                                   'payload': payload,
                                   'logs': []})
            self._depth += 1
        else:
            while not self._queue.empty():
                self._commands[-1]['logs'].append(self._queue.get())
            self._depth = max(self._depth - 1, 0)


def get_protocol_api(
        version: Union[str, APIVersion],
        bundled_labware: Dict[str, 'LabwareDefinition'] = None,
        bundled_data: Dict[str, bytes] = None,
        extra_labware: Dict[str, 'LabwareDefinition'] = None,
        hardware_simulator: HardwareToManage = None)\
        -> protocol_api.ProtocolContext:
    """
    Build and return a :py:class:`ProtocolContext` connected to
    Virtual Smoothie.

    This can be used to run protocols from interactive Python sessions
    such as Jupyter or an interpreter on the command line:

    .. code-block:: python

        >>> from opentrons.simulate import get_protocol_api
        >>> protocol = get_protocol_api('2.0')
        >>> instr = protocol.load_instrument('p300_single', 'right')
        >>> instr.home()

    If ``extra_labware`` is not specified, any labware definitions saved in
    the ``labware`` directory of the Jupyter notebook directory will be
    available.

    :param version: The API version to use. This must be lower than
                    :py:attr:`opentrons.protocol_api.MAX_SUPPORTED_VERSION`.
                    It may be specified either as a string (``'2.0'``) or
                    as a :py:class:`.protocols.types.APIVersion`
                    (``APIVersion(2, 0)``).
    :param bundled_labware: If specified, a mapping from labware names to
                            labware definitions for labware to consider in the
                            protocol. Note that if you specify this, _only_
                            labware in this argument will be allowed in the
                            protocol. This is preparation for a beta feature
                            and is best not used.
    :param bundled_data: If specified, a mapping from filenames to contents
                         for data to be available in the protocol from
                         :py:attr:`.ProtocolContext.bundled_data`.
    :param extra_labware: If specified, a mapping from labware names to
                          labware definitions for labware to consider in the
                          protocol in addition to those stored on the robot.
                          If this is an empty dict, and this function is called
                          on a robot, it will look in the 'labware'
                          subdirectory of the Jupyter data directory for
                          custom labware.
    :param hardware_simulator: If specified, a hardware simulator instance.
    :returns opentrons.protocol_api.ProtocolContext: The protocol context.
    """
    if isinstance(version, str):
        checked_version = parse.version_from_string(version)
    elif not isinstance(version, APIVersion):
        raise TypeError('version must be either a string or an APIVersion')
    else:
        checked_version = version
    if extra_labware is None\
       and IS_ROBOT\
       and JUPYTER_NOTEBOOK_LABWARE_DIR.is_dir():  # type: ignore
        extra_labware = labware_from_paths(
            [str(JUPYTER_NOTEBOOK_LABWARE_DIR)])
    return _build_protocol_context(
        checked_version, bundled_labware, bundled_data,
        extra_labware, hardware_simulator)


def _build_protocol_context(
        version: APIVersion = None,
        bundled_labware: Dict[str, 'LabwareDefinition'] = None,
        bundled_data: Dict[str, bytes] = None,
        extra_labware: Dict[str, 'LabwareDefinition'] = None,
        hardware_simulator: HardwareToManage = None,)\
        -> protocol_api.ProtocolContext:
    """ Internal version of :py:meth:`get_protocol_api` that allows deferring
    version specification for use with
    :py:meth:`.protocol_api.execute.run_protocol`
    """
    ctx_impl = ProtocolContextImplementation(
        bundled_labware=bundled_labware,
        bundled_data=bundled_data,
        api_version=version,
        extra_labware=extra_labware,
        hardware=hardware_simulator
    )
    context = protocol_api.contexts.ProtocolContext(
        implementation=ctx_impl,
        api_version=version
    )
    context.home()
    return context


def bundle_from_sim(
        protocol: PythonProtocol,
        context: opentrons.protocol_api.ProtocolContext)\
        -> BundleContents:
    """
    From a protocol, and the context that has finished simulating that
    protocol, determine what needs to go in a bundle for the protocol.
    """
    bundled_labware: Dict[str, 'LabwareDefinition'] = {}
    for lw in context.loaded_labwares.values():
        if isinstance(lw, opentrons.protocol_api.labware.Labware)\
           and lw.uri not in bundled_labware:
            bundled_labware[lw.uri] = lw._implementation.get_definition()

    return BundleContents(protocol.text,
                          bundled_data=context.bundled_data,
                          bundled_labware=bundled_labware,
                          bundled_python={})


def simulate(protocol_file: TextIO,
             file_name: str = None,
             custom_labware_paths: List[str] = None,
             custom_data_paths: List[str] = None,
             propagate_logs: bool = False,
             hardware_simulator_file_path: str = None,
             log_level: str = 'warning') -> Tuple[List[Mapping[str, Any]],
                                                  Optional[BundleContents]]:
    """
    Simulate the protocol itself.

    This is a one-stop function to simulate a protocol, whether python or json,
    no matter the api version, from external (i.e. not bound up in other
    internal server infrastructure) sources.

    To simulate an opentrons protocol from other places, pass in a file like
    object as protocol_file; this function either returns (if the simulation
    has no problems) or raises an exception.

    To call from the command line use either the autogenerated entrypoint
    ``opentrons_simulate`` (``opentrons_simulate.exe``, on windows) or
    ``python -m opentrons.simulate``.

    The return value is the run log, a list of dicts that represent the
    commands executed by the robot; and either the contents of the protocol
    that would be required to bundle, or ``None``.

    Each dict element in the run log has the following keys:

        - ``level``: The depth at which this command is nested - if this an
                     aspirate inside a mix inside a transfer, for instance,
                     it would be 3.
        - ``payload``: The command, its arguments, and how to format its text.
                       For more specific details see
                       :py:mod:`opentrons.commands`. To format a message from
                       a payload do ``payload['text'].format(**payload)``.
        - ``logs``: Any log messages that occurred during execution of this
                    command, as a logging.LogRecord

    :param file-like protocol_file: The protocol file to simulate.
    :param str file_name: The name of the file
    :param custom_labware_paths: A list of directories to search for custom
                                 labware, or None. Ignored if the apiv2 feature
                                 flag is not set. Loads valid labware from
                                 these paths and makes them available to the
                                 protocol context.
    :param custom_data_paths: A list of directories or files to load custom
                              data files from. Ignored if the apiv2 feature
                              flag if not set. Entries may be either files or
                              directories. Specified files and the
                              non-recursive contents of specified directories
                              are presented by the protocol context in
                              :py:attr:`.ProtocolContext.bundled_data`.
    :param hardware_simulator_file_path: A path to a JSON file defining a
                                         hardware simulator.
    :param propagate_logs: Whether this function should allow logs from the
                           Opentrons stack to propagate up to the root handler.
                           This can be useful if you're integrating this
                           function in a larger application, but most logs that
                           occur during protocol simulation are best associated
                           with the actions in the protocol that cause them.
                           Default: ``False``
    :type propagate_logs: bool
    :param log_level: The level of logs to capture in the runlog. Default:
                      ``'warning'``
    :type log_level: 'debug', 'info', 'warning', or 'error'
    :returns: A tuple of a run log for user output, and possibly the required
              data to write to a bundle to bundle this protocol. The bundle is
              only emitted if bundling is allowed (see
              :py:meth:`allow_bundling`)  and this is an unbundled Protocol API
              v2 python protocol. In other cases it is None.
    """
    stack_logger = logging.getLogger('opentrons')
    stack_logger.propagate = propagate_logs

    contents = protocol_file.read()
    if custom_labware_paths:
        extra_labware = labware_from_paths(custom_labware_paths)
    else:
        extra_labware = {}

    if custom_data_paths:
        extra_data = datafiles_from_paths(custom_data_paths)
    else:
        extra_data = {}

    hardware_simulator = None
    if hardware_simulator_file_path:
        hardware_simulator = asyncio.get_event_loop().run_until_complete(
            load_simulator(pathlib.Path(hardware_simulator_file_path))
        )

    protocol = parse.parse(contents, file_name,
                           extra_labware=extra_labware,
                           extra_data=extra_data)
    bundle_contents:  Optional[BundleContents] = None

    if getattr(protocol, 'api_level', APIVersion(2, 0)) < APIVersion(2, 0):
        def _simulate_v1():
            opentrons.robot.disconnect()
            opentrons.robot.reset()
            scraper = CommandScraper(stack_logger, log_level,
                                     opentrons.robot.broker)
            exec(protocol.contents, {})  # type: ignore
            return scraper

        scraper = _simulate_v1()
    else:
        # we want a None literal rather than empty dict so get_protocol_api
        # will look for custom labware if this is a robot
        gpa_extras = getattr(protocol, 'extra_labware', None) or None
        context = get_protocol_api(
            getattr(protocol, 'api_level', MAX_SUPPORTED_VERSION),
            bundled_labware=getattr(protocol, 'bundled_labware', None),
            bundled_data=getattr(protocol, 'bundled_data', None),
            hardware_simulator=hardware_simulator,
            extra_labware=gpa_extras)
        broker = context.broker
        scraper = CommandScraper(stack_logger,
                                 log_level,
                                 broker)
        try:
            execute.run_protocol(protocol, context)
            if isinstance(protocol, PythonProtocol)\
               and protocol.api_level >= APIVersion(2, 0)\
               and protocol.bundled_labware is None\
               and allow_bundle():
                bundle_contents = bundle_from_sim(
                    protocol, context)
        finally:
            context.cleanup()

    return scraper.commands, bundle_contents


def format_runlog(runlog: List[Mapping[str, Any]]) -> str:
    """
    Format a run log (return value of :py:meth:`simulate``) into a
    human-readable string

    :param runlog: The output of a call to :py:func:`simulate`
    """
    to_ret = []
    for command in runlog:
        to_ret.append(
            '\t' * command['level']
            + command['payload'].get('text', '').format(**command['payload']))
        if command['logs']:
            to_ret.append('\t' * command['level'] + 'Logs from this command:')
            to_ret.extend(
                ['\t' * command['level']
                 + f'{l.levelname} ({l.module}): {l.msg}' % l.args
                 for l in command['logs']])  # noqa: E741
    return '\n'.join(to_ret)


def _get_bundle_args(
        parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        '-b', '--bundle', nargs='?', const='PROTOCOL.ot2.zip', default=None,
        action='store', type=str,
        help='Bundle the specified protocol file, any labware used in it, and '
             'any files in the data directories specified with -D into a '
             'bundle. This bundle can be executed on a robot and carries with '
             'it all the custom labware and data required to run. Without a '
             'value specified in this argument, the bundle will be called '
             '(protocol name without the .py).ot2.zip, but you can specify '
             'a different output name. \n'
             'These bundles are a beta feature, and their behavior may change')
    return parser


def allow_bundle() -> bool:
    """
    Check if bundling is allowed with a special not-exposed-to-the-app flag.

    Returns ``True`` if the environment variable
    ``OT_API_FF_allowBundleCreation`` is ``"1"``
    """
    return os.getenv('OT_API_FF_allowBundleCreation') == '1'


def get_arguments(
        parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """ Get the argument parser for this module

    Useful if you want to use this module as a component of another CLI program
    and want to add its arguments.

    :param parser: A parser to add arguments to. If not specified, one will be
                   created.
    :returns argparse.ArgumentParser: The parser with arguments added.
    """
    parser.add_argument(
        '-l', '--log-level',
        choices=['debug', 'info', 'warning', 'error', 'none'],
        default='warning',
        help='Specify the level filter for logs to show on the command line. '
        'Log levels below warning can be chatty. If "none", do not show logs')

    parser.add_argument(
        '-L', '--custom-labware-path',
        action='append', default=[os.getcwd()],
        help='Specify directories to search for custom labware definitions. '
             'You can specify this argument multiple times. Once you specify '
             'a directory in this way, labware definitions in that directory '
             'will become available in ProtocolContext.load_labware(). '
             'Only directories specified directly by '
             'this argument are searched, not their children. JSON files that '
             'do not define labware will be ignored with a message. '
             'By default, the current directory (the one from which you are '
             'invoking this program) will be searched for labware.')
    parser.add_argument(
        '-D', '--custom-data-path',
        action='append', nargs='?', const='.', default=[],
        help='Specify directories to search for custom data files. '
             'You can specify this argument multiple times. Once you specify '
             'a directory in this way, files located in the specified '
             'directory will be available in ProtocolContext.bundled_data. '
             'Note that bundle execution will still only allow data files in '
             'the bundle. If you specify this without a path, it will '
             'add the current path implicitly. If you do not specify this '
             'argument at all, no data files will be added. Any file in the '
             'specified paths will be loaded into memory and included in the '
             'bundle if --bundle is passed, so be careful that any directory '
             'you specify has only the files you want. It is usually a '
             'better idea to use -d so no files are accidentally included. '
             'Also note that data files are made available as their name, not '
             'their full path, so name them uniquely.')
    parser.add_argument(
        '-s', '--custom-hardware-simulator-file',
        type=str, default=None,
        help='Specify a file that describes the features present in the '
             'hardware simulator. Features can be instruments, modules, and '
             'configuration.')
    parser.add_argument(
        '-d', '--custom-data-file',
        action='append', default=[],
        help='Specify data files to be made available in '
             'ProtocolContext.bundled_data (and possibly bundled if --bundle '
             'is passed). Can be specified multiple times with different '
             'files. It is usually a better idea to use this than -D because '
             'there is less possibility of accidentally including something.')
    if allow_bundle():
        parser = _get_bundle_args(parser)

    parser.add_argument(
        'protocol', metavar='PROTOCOL',
        type=argparse.FileType('rb'),
        help='The protocol file to simulate. If you pass \'-\', you can pipe '
        'the protocol via stdin; this could be useful if you want to use this '
        'utility as part of an automated workflow.')
    parser.add_argument(
        '-v', '--version', action='version',
        version=f'%(prog)s {opentrons.__version__}',
        help='Print the opentrons package version and exit')
    parser.add_argument(
        '-o', '--output', action='store',
        help='What to output during simulations',
        choices=['runlog', 'nothing'],
        default='runlog')
    return parser


def _get_bundle_dest(
        bundle_name: Optional[str],
        default_key: str,
        proto_name: str) -> Optional[BinaryIO]:
    if bundle_name == default_key:
        protopath = pathlib.Path(proto_name)
        # strip all the suffixes since protocols are often named
        # .ot2.zip
        if protopath.name.endswith('.ot2.py'):
            protoname = pathlib.Path(protopath.stem).stem
        else:
            protoname = protopath.stem
        bundle_name = str((pathlib.Path.cwd() / protoname)
                          .with_suffix('.ot2.zip'))
        return open(bundle_name, 'wb')
    elif bundle_name:
        return open(bundle_name, 'wb')
    else:
        return None


# Note - this script is also set up as a setuptools entrypoint and thus does
# an absolute minimum of work since setuptools does something odd generating
# the scripts
def main() -> int:
    """ Run the simulation """
    parser = argparse.ArgumentParser(prog='opentrons_simulate',
                                     description='Simulate an OT-2 protocol')
    parser = get_arguments(parser)

    args = parser.parse_args()
    # Try to migrate api v1 containers if needed

    runlog, maybe_bundle = simulate(
        args.protocol,
        args.protocol.name,
        getattr(args, 'custom_labware_path', []),
        getattr(args, 'custom_data_path', [])
        + getattr(args, 'custom_data_file', []),
        hardware_simulator_file_path=getattr(args,
                                             'custom_hardware_simulator_file'),
        log_level=args.log_level)

    if maybe_bundle:
        bundle_name = getattr(args, 'bundle', None)
        if bundle_name == args.protocol.name:
            raise RuntimeError(
                'Bundle path and input path must be different')
        bundle_dest = _get_bundle_dest(
            bundle_name, 'PROTOCOL.ot2.zip', args.protocol.name)
        if bundle_dest:
            bundle.create_bundle(maybe_bundle, bundle_dest)

    if args.output == 'runlog':
        print(format_runlog(runlog))

    return 0


if __name__ == '__main__':
    sys.exit(main())



"""
This was initially generated by datamodel-codegen from the protocol schema in
shared-data. It's been modified by hand to be more friendly.

TODO: 20210330 Amit - consider moving this to opentrons-shared-data.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra, Field
from typing_extensions import Literal

from .labware_definition import LabwareDefinition as LabwareDefinition
from opentrons_shared_data.protocol import dev_types

CommandAspirate: dev_types.AspirateCommandId = 'aspirate'
CommandDispense: dev_types.DispenseCommandId = 'dispense'
CommandAirGap: dev_types.AirGapCommandId = 'airGap'
CommandBlowout: dev_types.BlowoutCommandId = 'blowout'
CommandTouchTip: dev_types.TouchTipCommandId = 'touchTip'
CommandPickUpTip: dev_types.PickUpTipCommandId = 'pickUpTip'
CommandDropTip: dev_types.DropTipCommandId = 'dropTip'
CommandMoveToSlot: dev_types.MoveToSlotCommandId = 'moveToSlot'
CommandMoveToWell: dev_types.MoveToWellCommandId = 'moveToWell'
CommandDelay: dev_types.DelayCommandId = 'delay'
CommandMagneticModuleEngage: \
    dev_types.MagneticModuleEngageCommandId = 'magneticModule/engageMagnet'
CommandMagneticModuleDisengage:\
    dev_types.MagneticModuleDisengageCommandId = 'magneticModule/disengageMagnet'
CommandTemperatureModuleSetTarget: \
    dev_types.TemperatureModuleSetTargetCommandId = \
    'temperatureModule/setTargetTemperature'
CommandTemperatureModuleAwait: \
    dev_types.TemperatureModuleAwaitCommandId = 'temperatureModule/awaitTemperature'
CommandTemperatureModuleDeactivate: \
    dev_types.TemperatureModuleDeactivateCommandId = 'temperatureModule/deactivate'
CommandThermocyclerSetTargetBlock: \
    dev_types.ThermocyclerSetTargetBlockCommandId = 'thermocycler/setTargetBlockTemperature'  # noqa: E501
CommandThermocyclerSetTargetLid: \
    dev_types.ThermocyclerSetTargetLidCommandId = 'thermocycler/setTargetLidTemperature'  # noqa: E501
CommandThermocyclerAwaitLidTemperature:\
    dev_types.ThermocyclerAwaitLidTemperatureCommandId = 'thermocycler/awaitLidTemperature'  # noqa: E501
CommandThermocyclerAwaitBlockTemperature:\
    dev_types.ThermocyclerAwaitBlockTemperatureCommandId = 'thermocycler/awaitBlockTemperature'  # noqa: E501
CommandThermocyclerDeactivateBlock:\
    dev_types.ThermocyclerDeactivateBlockCommandId = 'thermocycler/deactivateBlock'
CommandThermocyclerDeactivateLid:\
    dev_types.ThermocyclerDeactivateLidCommandId = 'thermocycler/deactivateLid'
CommandThermocyclerOpenLid:\
    dev_types.ThermocyclerOpenLidCommandId = 'thermocycler/openLid'
CommandThermocyclerCloseLid:\
    dev_types.ThermocyclerCloseLidCommandId = 'thermocycler/closeLid'
CommandThermocyclerRunProfile:\
    dev_types.ThermocyclerRunProfileCommandId = 'thermocycler/runProfile'
CommandThermocyclerAwaitProfile:\
    dev_types.ThermocyclerAwaitProfileCommandId = 'thermocycler/awaitProfileComplete'


class Metadata(BaseModel):
    """
    Optional metadata about the protocol
    """

    protocolName: Optional[str] = Field(
        None, description='A short, human-readable name for the protocol'
    )
    author: Optional[str] = Field(
        None, description='The author or organization who created the protocol'
    )
    description: Optional[Optional[str]] = Field(
        None, description='A text description of the protocol.'
    )
    created: Optional[float] = Field(
        None, description='UNIX timestamp when this file was created'
    )
    lastModified: Optional[Optional[float]] = Field(
        None, description='UNIX timestamp when this file was last modified'
    )
    category: Optional[Optional[str]] = Field(
        None, description='Category of protocol (eg, "Basic Pipetting")'
    )
    subcategory: Optional[Optional[str]] = Field(
        None, description='Subcategory of protocol (eg, "Cell Plating")'
    )
    tags: Optional[List[str]] = Field(
        None, description='Tags to be used in searching for this protocol'
    )


class DesignerApplication(BaseModel):
    """
    Optional data & metadata not required to execute the protocol, used by the
    application that created this protocol
    """

    name: Optional[str] = Field(
        None,
        description='Name of the application that created the protocol. Should '
                    'be namespaced under the organization or individual who '
                    'owns the organization, eg "opentrons/protocol-designer"',
    )
    version: Optional[str] = Field(
        None, description='Version of the application that created the protocol'
    )
    data: Optional[Dict[str, Any]] = Field(
        None,
        description='Any data used by the application that created this protocol'
    )


class Robot(BaseModel):
    model: Literal['OT-2 Standard'] = Field(
        ...,
        description='Model of the robot this protocol is written for '
                    '(currently only OT-2 Standard is supported)',
    )


class ModuleOnlyParams(BaseModel):
    module: str


class OffsetFromBottomMm(BaseModel):
    """
    Offset from bottom of well in millimeters
    """

    offsetFromBottomMm: float = Field(
        ..., description='Millimeters for pipette location offsets')


class PipetteAccessParams(BaseModel):
    pipette: str
    labware: str
    well: str


class VolumeParams(BaseModel):
    volume: float


class FlowRate(BaseModel):
    flowRate: float = Field(
        ..., description='Flow rate in uL/sec. Must be greater than 0', ge=0.0
    )


class Params2(PipetteAccessParams, OffsetFromBottomMm):
    pass


class Params1(Params2, FlowRate):
    pass


class Params(Params1, Params2, VolumeParams):
    pass


class LiquidCommand(BaseModel):
    """
    Aspirate / dispense / air gap commands
    """

    command: Literal['aspirate', 'dispense', 'airGap']
    params: Params


class BlowoutCommand(BaseModel):
    """
    Blowout command
    """

    command: Literal['blowout']
    params: Params1


class TouchTipCommand(BaseModel):
    """
    Touch tip commands
    """

    command: Literal['touchTip']
    params: Params2


class PickUpDropTipCommand(BaseModel):
    """
    Pick up tip / drop tip commands
    """

    command: Literal['pickUpTip', 'dropTip']
    params: PipetteAccessParams


class Offset(BaseModel):
    """
    Optional offset from slot bottom left corner, in mm
    """

    x: float
    y: float
    z: float


class Params3(BaseModel):
    pipette: str
    slot: str = Field(
        ...,
        description="string '1'-'12', or special string 'span7_8_10_11' signify "
                    "it's a slot on the OT-2 deck. If it's a UUID, it's the "
                    "slot on the module referenced by that ID.",
    )
    offset: Optional[Offset] = Field(
        None, description='Optional offset from slot bottom left corner, in mm'
    )
    minimumZHeight: Optional[float] = Field(
        None,
        description="Optional minimal Z margin in mm. If this is larger than "
                    "the API's default safe Z margin, it will make the arc "
                    "higher. If it's smaller, it will have no effect. Specifying "
                    "this for movements that would not arc (moving within the "
                    "same well in the same labware) will cause an arc movement "
                    "instead.",
        ge=0.0,
    )
    forceDirect: Optional[bool] = Field(
        None,
        description="Default is false. If true, moving from one labware/well to "
                    "another will not arc to the default safe z, but instead will "
                    "move directly to the specified location. This will also force "
                    "the 'minimumZHeight' param to be ignored. A 'direct' movement "
                    "is in X/Y/Z simultaneously",
    )


class MoveToSlotCommand(BaseModel):
    """
    Move to slot command. NOTE: this is an EXPERIMENTAL command, its behavior is
    subject to change in future releases.
    """

    command: Literal['moveToSlot']
    params: Params3


class Params4(BaseModel):
    wait: Union[float, Literal[True]] = Field(
        ...,
        description='either a number of seconds to wait (fractional values OK), '
                    'or `true` to wait indefinitely until the user manually '
                    'resumes the protocol',
    )
    message: Optional[str] = Field(
        None,
        description='optional message describing the delay'
    )


class DelayCommand(BaseModel):
    """
    Delay command
    """

    command: Literal['delay']
    params: Params4


class Params5(BaseModel):
    engageHeight: float = Field(
        ...,
        description='Height in mm(*) from bottom plane of labware (above if positive, '
                    'below if negative). *NOTE: for magneticModuleV1 (aka GEN1), these '
                    'are not true mm but an arbitrary unit equal to 0.5mm. So '
                    '`engageHeight: 2` means 1mm above the labware plane if the '
                    'command is for a GEN1 magnetic module, but would mean 2mm '
                    'above the labware plane for GEN2 module',
    )
    module: str


class MagneticModuleEngageCommand(BaseModel):
    """
    Magnetic module engage command. Engages magnet to specified height
    """

    command: Literal['magneticModule/engageMagnet']
    params: Params5


class MagneticModuleDisengageCommand(BaseModel):
    """
    Magnetic module disengage command. Moves magnet down to disengaged (home) position
    """

    command: Literal['magneticModule/disengageMagnet']
    params: ModuleOnlyParams


class Params6(BaseModel):
    module: str
    temperature: float


class TemperatureModuleSetTargetCommand(BaseModel):
    """
    Temperature module set target temperature command. Module will begin moving
    to the target temperature. This command is non-blocking, it does not delay
    until the temperature is reached.
    """

    command: Literal['temperatureModule/setTargetTemperature']
    params: Params6


class Params7(BaseModel):
    module: str
    temperature: float


class TemperatureModuleAwaitTemperatureCommand(BaseModel):
    """
    Temperature module await temperature command. Delay further protocol execution
    until the specified temperature is reached.
    """

    command: Literal['temperatureModule/awaitTemperature']
    params: Params7


class TemperatureModuleDeactivateCommand(BaseModel):
    """
    Temperature module deactivate command. Module will stop actively controlling
    its temperature and drift to ambient temperature.
    """

    command: Literal['temperatureModule/deactivate']
    params: ModuleOnlyParams


class Params8(BaseModel):
    module: str
    temperature: float
    volume: Optional[float] = None


class ThermocyclerSetTargetBlockTemperatureCommand(BaseModel):
    """
    Thermocycler set target block temperature command. Lid will begin moving to
    the target temperature. This command is non-blocking, it does not delay until
    the temperature is reached.
    """

    command: Literal['thermocycler/setTargetBlockTemperature']
    params: Params8


class Params9(BaseModel):
    module: str
    temperature: float


class ThermocyclerSetTargetLidTemperatureCommand(BaseModel):
    """
    Thermocycler set target lid temperature command. Lid will begin moving to the
    target temperature. This command is non-blocking, it does not delay until
    the temperature is reached.
    """

    command: Literal['thermocycler/setTargetLidTemperature']
    params: Params9


class Params10(BaseModel):
    module: str
    temperature: float


class ThermocyclerAwaitBlockTemperatureCommand(BaseModel):
    """
    Thermocycler await block temperature command. Delay further protocol execution
    until the specified temperature is reached.
    """

    command: Literal['thermocycler/awaitBlockTemperature']
    params: Params10


class Params11(BaseModel):
    module: str
    temperature: float


class ThermocyclerAwaitLidTemperatureCommand(BaseModel):
    """
    Thermocycler await lid temperature command. Delay further protocol execution
    until the specified temperature is reached.
    """

    command: Literal['thermocycler/awaitLidTemperature']
    params: Params11


class ThermocyclerDeactivateBlockCommand(BaseModel):
    """
    Thermocycler deactivate block command. Module will stop actively controlling
    its block temperature.
    """

    command: Literal['thermocycler/deactivateBlock']
    params: ModuleOnlyParams


class ThermocyclerDeactivateLidCommand(BaseModel):
    """
    Thermocycler deactivate lid command. Module will stop actively controlling
    its lid temperature.
    """

    command: Literal['thermocycler/deactivateLid']
    params: ModuleOnlyParams


class ThermocyclerOpenLidCommand(BaseModel):
    """
    Thermocycler open lid command. This command will block until the lid is
    fully open.
    """

    command: Literal['thermocycler/openLid']
    params: ModuleOnlyParams


class ThermocyclerCloseLidCommand(BaseModel):
    """
    Thermocycler close lid command. This command will block until the lid is
    fully closed.
    """

    command: Literal['thermocycler/closeLid']
    params: ModuleOnlyParams


class ProfileItem(BaseModel):
    temperature: float = Field(
        ...,
        description='Target temperature of profile step')
    holdTime: float = Field(
        ...,
        description='Time (in seconds) to hold once temperature is reached'
    )


class Params12(BaseModel):
    module: str
    profile: List[ProfileItem]
    volume: float


class ThermocyclerRunProfile(BaseModel):
    """
    Thermocycler run profile command. Begin running the specified profile steps on
    the thermocycler. This command is non-blocking, it does not delay protocol
    execution outside of thermocycler steps. No more thermocycler commands should
    be given until a 'thermocycler/awaitProfileComplete' command is executed.
    """

    command: Literal['thermocycler/runProfile']
    params: Params12


class ThermocyclerAwaitProfileCompleteCommand(BaseModel):
    """
    Thermocycler await profile complete command. Blocks further protocol
    execution until profile execution is complete.
    """

    command: Literal['thermocycler/awaitProfileComplete']
    params: ModuleOnlyParams


class Offset1(BaseModel):
    """
    Optional offset from well bottom center, in mm
    """

    x: float
    y: float
    z: float


class Params13(BaseModel):
    pipette: str
    labware: str
    well: str
    offset: Optional[Offset1] = Field(
        None,
        description='Optional offset from well bottom center, in mm'
    )
    minimumZHeight: Optional[float] = Field(
        None,
        description="Optional minimal Z margin in mm. If this is larger than "
                    "the API's default safe Z margin, it will make the arc higher. "
                    "If it's smaller, it will have no effect. Specifying this for "
                    "movements that would not arc (moving within the same well in "
                    "the same labware) will cause an arc movement instead.",
        ge=0.0,
    )
    forceDirect: Optional[bool] = Field(
        None,
        description="Default is false. If true, moving from one labware/well "
                    "to another will not arc to the default safe z, but instead "
                    "will move directly to the specified location. This will also "
                    "force the 'minimumZHeight' param to be ignored. A 'direct' "
                    "movement is in X/Y/Z simultaneously",
    )


class MoveToWellCommand(BaseModel):
    """
    Move to well command. Move the pipette's critical point to the specified well
    in a labware, with an optional offset. The pipette's critical point is a
    reference point on the pipette. The critical point can be one of the following:
    (1) Single-channel pipette with no tip: end of nozzle. (2) Multi-channel
    pipette with no tip: end of backmost nozzle. (3) Single-channel pipette with
    a tip: end of tip. (4) Multi-channel pipette with tip: end of tip on backmost
    nozzle.
    """

    command: Literal['moveToWell']
    params: Params13


AllCommands = Union[
    LiquidCommand,
    BlowoutCommand,
    TouchTipCommand,
    PickUpDropTipCommand,
    MoveToSlotCommand,
    DelayCommand,
    MagneticModuleEngageCommand,
    MagneticModuleDisengageCommand,
    TemperatureModuleSetTargetCommand,
    TemperatureModuleAwaitTemperatureCommand,
    TemperatureModuleDeactivateCommand,
    ThermocyclerSetTargetBlockTemperatureCommand,
    ThermocyclerSetTargetLidTemperatureCommand,
    ThermocyclerAwaitBlockTemperatureCommand,
    ThermocyclerAwaitLidTemperatureCommand,
    ThermocyclerDeactivateBlockCommand,
    ThermocyclerDeactivateLidCommand,
    ThermocyclerOpenLidCommand,
    ThermocyclerCloseLidCommand,
    ThermocyclerRunProfile,
    ThermocyclerAwaitProfileCompleteCommand,
    MoveToWellCommand,
]


class Pipettes(BaseModel):
    """
    Fields describing an individual pipette
    """

    class Config:
        extra = Extra.allow

    mount: Literal['left', 'right'] = Field(
        ..., description='Where the pipette is mounted'
    )
    name: str = Field(
        ...,
        description='Name of a pipette. Does not contain info about specific '
                    'model/version. Should match keys in pipetteNameSpecs.json file',
    )


class Labware(BaseModel):
    """
    Fields describing a single labware on the deck
    """

    class Config:
        extra = Extra.allow

    slot: str = Field(
        ...,
        description="string '1'-'12', or special string 'span7_8_10_11' signify "
                    "it's a slot on the OT-2 deck. If it's a UUID, it's the "
                    "slot on the module referenced by that ID.",
    )
    definitionId: str = Field(
        ...,
        description='reference to this labware\'s ID in "labwareDefinitions"'
    )
    displayName: Optional[str] = Field(
        None,
        description='An optional human-readable nickname for this labware. '
                    'Eg "Buffer Trough"',
    )


class Modules(BaseModel):
    """
    Fields describing a single module on the deck
    """

    class Config:
        extra = Extra.allow

    slot: str = Field(
        ...,
        description="string '1'-'12', or special string 'span7_8_10_11' signify "
                    "it's a slot on the OT-2 deck. If it's a UUID, it's the "
                    "slot on the module referenced by that ID.",
    )
    model: str = Field(
        ...,
        description="model of module. Eg 'magneticModuleV1' or 'magneticModuleV2'. "
                    "This should match a top-level key in "
                    "shared-data/module/definitions/2.json",
    )


class Model(BaseModel):
    otSharedSchema: Optional[
        Literal['#/protocol/schemas/5', '#/protocol/schemas/4']
    ] = Field(
        None,
        alias='$otSharedSchema',
        description='The path to a valid Opentrons shared schema relative to '
                    'the shared-data directory, without its extension.',
    )
    schemaVersion: Literal[1, 2, 3, 4, 5] = Field(
        ..., description='Schema version of a protocol is a single integer'
    )
    metadata: Metadata = Field(
        ...,
        description='Optional metadata about the protocol'
    )
    designerApplication: Optional[DesignerApplication] = Field(
        None,
        description='Optional data & metadata not required to execute the protocol, '
                    'used by the application that created this protocol',
    )
    robot: Robot
    pipettes: Dict[str, Pipettes] = Field(
        ...,
        description='The pipettes used in this protocol, keyed by an arbitrary '
                    'unique ID',
    )
    labwareDefinitions: Dict[str, LabwareDefinition] = Field(
        ...,
        description='All labware definitions used by labware in this protocol, '
                    'keyed by UUID',
    )
    labware: Dict[str, Labware] = Field(
        ...,
        description='All types of labware used in this protocol, and references '
                    'to their definitions',
    )
    modules: Optional[Dict[str, Modules]] = Field(
        None, description='All modules used in this protocol'
    )
    commands: List[AllCommands] = Field(
        None,
        description="An array of command objects representing steps to be executed "
                    "on the robot"
    )
    commandAnnotations: Optional[Dict[str, Any]] = Field(
        None,
        description='An optional object of annotations associated with commands. '
                    'Its usage has not yet been defined, so its shape is not '
                    'enforced by this schema.',
    )



"""Protocol engine state management."""
from __future__ import annotations
from typing import List, Sequence

from opentrons_shared_data.deck.dev_types import DeckDefinitionV2

from .. import commands as cmd
from ..resources import DeckFixedLabware
from .substore import CommandReactive
from .commands import CommandStore, CommandState
from .labware import LabwareStore, LabwareState
from .pipettes import PipetteStore, PipetteState
from .geometry import GeometryStore, GeometryState
from .motion import MotionStore, MotionState


class StateView:
    """A read-only view of a StateStore."""

    _command_store: CommandStore
    _labware_store: LabwareStore
    _pipette_store: PipetteStore
    _geometry_store: GeometryStore
    _motion_store: MotionStore

    def __init__(
        self,
        command_store: CommandStore,
        labware_store: LabwareStore,
        pipette_store: PipetteStore,
        geometry_store: GeometryStore,
        motion_store: MotionStore,
    ) -> None:
        """Initialize a StateView."""
        self._command_store = command_store
        self._labware_store = labware_store
        self._pipette_store = pipette_store
        self._geometry_store = geometry_store
        self._motion_store = motion_store

    @classmethod
    def create_view(cls, target: StateStore) -> StateView:
        """Create a read-only view of a target StateStore."""
        return cls(
            command_store=target._command_store,
            labware_store=target._labware_store,
            pipette_store=target._pipette_store,
            geometry_store=target._geometry_store,
            motion_store=target._motion_store,
        )

    @property
    def commands(self) -> CommandState:
        """Get commands sub-state."""
        return self._command_store.state

    @property
    def labware(self) -> LabwareState:
        """Get labware sub-state."""
        return self._labware_store.state

    @property
    def pipettes(self) -> PipetteState:
        """Get pipettes sub-state."""
        return self._pipette_store.state

    @property
    def geometry(self) -> GeometryState:
        """Get geometry sub-state."""
        return self._geometry_store.state

    @property
    def motion(self) -> MotionState:
        """Get motion sub-state."""
        return self._motion_store.state


class StateStore(StateView):
    """ProtocolEngine state store.

    A StateStore manages several substores, which will modify themselves in
    reaction to commands and other protocol events. Only Store classes should
    be allowed to modify State classes.
    """

    def __init__(
        self,
        deck_definition: DeckDefinitionV2,
        deck_fixed_labware: Sequence[DeckFixedLabware],
    ) -> None:
        """Initialize a StateStore."""
        command_store = CommandStore()
        labware_store = LabwareStore(
            deck_fixed_labware=deck_fixed_labware
        )
        pipette_store = PipetteStore()
        geometry_store = GeometryStore(
            deck_definition=deck_definition,
            labware_store=labware_store,
        )
        motion_store = MotionStore(
            labware_store=labware_store,
            pipette_store=pipette_store,
            geometry_store=geometry_store,
        )

        # attach stores to self via StateView constructor
        super().__init__(
            command_store=command_store,
            labware_store=labware_store,
            pipette_store=pipette_store,
            geometry_store=geometry_store,
            motion_store=motion_store
        )

        self._lifecycle_substores: List[CommandReactive] = [
            labware_store,
            pipette_store,
            geometry_store,
            motion_store,
        ]

    def handle_command(
        self,
        command: cmd.CommandType,
        command_id: str
    ) -> None:
        """Modify State in reaction to a Command."""
        self._command_store.handle_command(command, command_id)

        if isinstance(command, cmd.CompletedCommand):
            for substore in self._lifecycle_substores:
                substore.handle_completed_command(command)



""" opentrons.util.entrypoint_util: functions common to entrypoints
"""

import logging
from json import JSONDecodeError
import pathlib
from typing import Dict, List, TYPE_CHECKING

from jsonschema import ValidationError  # type: ignore

from opentrons.protocol_api import labware
from opentrons.calibration_storage import helpers

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition
log = logging.getLogger(__name__)


def labware_from_paths(paths: List[str]) -> Dict[str, 'LabwareDefinition']:
    labware_defs: Dict[str, 'LabwareDefinition'] = {}

    for strpath in paths:
        log.info(f"local labware: checking path {strpath}")
        purepath = pathlib.PurePath(strpath)
        if purepath.is_absolute():
            path = pathlib.Path(purepath)
        else:
            path = pathlib.Path.cwd() / purepath
        if not path.is_dir():
            raise RuntimeError(f'{path} is not a directory')
        for child in path.iterdir():
            if child.is_file() and child.suffix.endswith('json'):
                try:
                    defn = labware.verify_definition(child.read_bytes())
                except (ValidationError, JSONDecodeError) as e:
                    log.info(f"{child}: invalid labware, ignoring")
                    log.debug(f"{child}: labware invalid because: {str(e)}")
                else:
                    uri = helpers.uri_from_definition(defn)
                    labware_defs[uri] = defn
                    log.info(f'loaded labware {uri} from {child}')
            else:
                log.info(f'ignoring {child} in labware path')
    return labware_defs


def datafiles_from_paths(paths: List[str]) -> Dict[str, bytes]:
    datafiles: Dict[str, bytes] = {}
    for strpath in paths:
        log.info(f"data files: checking path {strpath}")
        purepath = pathlib.PurePath(strpath)
        if purepath.is_absolute():
            path = pathlib.Path(purepath)
        else:
            path = pathlib.Path.cwd() / purepath
        if path.is_file():
            datafiles[path.name] = path.read_bytes()
            log.info(f'read {path} into custom data as {path.name}')
        elif path.is_dir():
            for child in path.iterdir():
                if child.is_file():
                    datafiles[child.name] = child.read_bytes()
                    log.info(f'read {child} into data path as {child.name}')
                else:
                    log.info(f'ignoring {child} in data path')
    return datafiles
